import os

import unittest
from unittest.mock import patch, Mock, call

import numpy as np
import scipy.sparse as sp

import openTSNE.affinity
from Orange.data import DiscreteVariable, ContinuousVariable, Domain, Table
from Orange.distance import Euclidean
from Orange.misc import DistMatrix
from Orange.preprocess import Normalize
from Orange.projection import manifold, TSNE
from Orange.projection.manifold import TSNEModel
from Orange.widgets.tests.base import (
    WidgetTest, WidgetOutputsTestMixin, ProjectionWidgetTestMixin
)
from Orange.widgets.tests.utils import simulate
from Orange.widgets.unsupervised.owtsne import OWtSNE, TSNERunner, Task, prepare_tsne_obj


class DummyTSNE(manifold.TSNE):
    def compute_affinities(self, X):

        class DummyAffinities(openTSNE.affinity.Affinities):
            def __init__(self, data=None, *args, **kwargs):
                n_samples = data.shape[0]
                self.P = sp.random(n_samples, n_samples, density=0.1)
                self.P /= self.P.sum()

            def to_new(self, data, return_distances=False):
                ones = np.ones((len(data), 2), float)
                if return_distances:
                    return ones, ones
                return ones

        return DummyAffinities(X)

    def compute_initialization(self, X):
        return np.ones((X.shape[0], 2), float)

    def fit(self, X, Y=None):
        return np.ones((X.shape[0], 2), float)


class DummyTSNEModel(manifold.TSNEModel):
    def transform(self, X, **kwargs):
        return np.ones((X.shape[0], 2), float)

    def optimize(self, n_iter, **kwargs):
        return self


class TestOWtSNE(WidgetTest, ProjectionWidgetTestMixin, WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        WidgetOutputsTestMixin.init(cls)
        cls.same_input_output_domain = False

        cls.signal_name = OWtSNE.Inputs.data
        cls.signal_data = cls.data

        cls.iris = Table("iris")
        cls.iris_distances = Euclidean(cls.iris)
        cls.housing = Table("housing")[:200]
        cls.housing_distances = Euclidean(cls.housing)

        # Load distance-only DistMatrix, without accompanying `.row_items`
        my_dir = os.path.dirname(__file__)
        datasets_dir = os.path.join(my_dir, '..', '..', '..', 'datasets')
        cls.datasets_dir = os.path.realpath(datasets_dir)
        cls.towns = DistMatrix.from_file(
            os.path.join(cls.datasets_dir, "slovenian-towns.dst")
        )

    def setUp(self):
        # For almost all the tests, we won't need to verify t-SNE validity and
        # the tests will run much faster if we dummy them out
        self.tsne = patch("Orange.projection.manifold.TSNE", new=DummyTSNE)
        self.tsne_model = patch("Orange.projection.manifold.TSNEModel", new=DummyTSNEModel)
        self.tsne.start()
        self.tsne_model.start()

        self.widget = self.create_widget(OWtSNE, stored_settings={"multiscale": False})

        self.class_var = DiscreteVariable("Stage name", values=("STG1", "STG2"))
        self.attributes = [ContinuousVariable("GeneName" + str(i)) for i in range(5)]
        self.domain = Domain(self.attributes, class_vars=self.class_var)
        self.empty_domain = Domain([], class_vars=self.class_var)

    def tearDown(self):
        # Some tests may not wait for the widget to finish, and the patched
        # methods might be unpatched before the widget finishes, resulting in
        # a very confusing crash.
        self.widget.onDeleteWidget()
        try:
            self.restore_mocked_functions()
        # If `restore_mocked_functions` was called in the test itself, stopping
        # the patchers here will raise a RuntimeError
        except RuntimeError as e:
            if str(e) != "stop called on unstarted patcher":
                raise e

    def restore_mocked_functions(self):
        self.tsne.stop()
        self.tsne_model.stop()

    def test_wrong_input(self):
        # no data
        data = None
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)

        # <2 rows
        data = Table.from_list(self.domain, [[1, 2, 3, 4, 5, 'STG1']])
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.not_enough_rows.is_shown())

        # no attributes
        data = Table.from_list(self.empty_domain, [['STG1']] * 2)
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.not_enough_cols.is_shown())

        # one attributes
        data = Table.from_list(self.empty_domain, [[1, 'STG1'],
                                                   [2, 'STG1']])
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.not_enough_cols.is_shown())

        # constant data
        data = Table.from_list(self.domain, [[1, 2, 3, 4, 5, 'STG1']] * 2)
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.constant_data.is_shown())

        # correct input
        data = Table.from_list(self.domain, [[1, 2, 3, 4, 5, 'STG1'],
                                             [5, 4, 3, 2, 1, 'STG1']])
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNotNone(self.widget.data)
        self.assertFalse(self.widget.Error.not_enough_rows.is_shown())
        self.assertFalse(self.widget.Error.not_enough_cols.is_shown())
        self.assertFalse(self.widget.Error.constant_data.is_shown())

    def test_input(self):
        data = Table.from_list(self.domain, [[1, 1, 1, 1, 1, 'STG1'],
                                             [2, 2, 2, 2, 2, 'STG1'],
                                             [4, 4, 4, 4, 4, 'STG2'],
                                             [5, 5, 5, 5, 5, 'STG2']])

        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()

    def test_attr_models(self):
        """Check possible values for 'Color', 'Shape', 'Size' and 'Label'"""
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        controls = self.widget.controls
        for var in self.data.domain.class_vars + self.data.domain.metas:
            self.assertIn(var, controls.attr_color.model())
            self.assertIn(var, controls.attr_label.model())
            if var.is_continuous:
                self.assertIn(var, controls.attr_size.model())
                self.assertNotIn(var, controls.attr_shape.model())
            if var.is_discrete:
                self.assertNotIn(var, controls.attr_size.model())
                self.assertIn(var, controls.attr_shape.model())

    def test_multiscale_changed_updates_ui(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertFalse(self.widget.controls.multiscale.isChecked())
        self.assertTrue(self.widget.perplexity_spin.isEnabled())
        self.widget.controls.multiscale.setChecked(True)
        self.assertFalse(self.widget.perplexity_spin.isEnabled())

        settings = self.widget.settingsHandler.pack_data(self.widget)
        w = self.create_widget(OWtSNE, stored_settings=settings)
        self.send_signal(w.Inputs.data, self.data, widget=w)
        self.assertTrue(w.controls.multiscale.isChecked())
        self.assertFalse(w.perplexity_spin.isEnabled())
        w.onDeleteWidget()

    def test_normalize_data(self):
        # Normalization should be checked by default
        self.assertTrue(self.widget.controls.normalize.isChecked())
        with patch("Orange.preprocess.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, self.data)
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            self.wait_until_finished()
            normalize.assert_called_once()

        # Disable checkbox
        self.widget.controls.normalize.setChecked(False)
        self.assertFalse(self.widget.controls.normalize.isChecked())
        with patch("Orange.preprocess.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, self.data)
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            self.wait_until_finished()
            normalize.assert_not_called()

        # Normalization shouldn't work on sparse data
        self.widget.controls.normalize.setChecked(True)
        self.assertTrue(self.widget.controls.normalize.isChecked())

        sparse_data = self.data.to_sparse()
        with patch("Orange.preprocess.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, sparse_data)
            self.assertFalse(self.widget.controls.normalize.isEnabled())
            self.wait_until_finished()
            normalize.assert_not_called()

    @patch("Orange.projection.manifold.TSNEModel.optimize")
    def test_exaggeration_is_passed_through_properly(self, optimize):
        def _check_exaggeration(call, exaggeration):
            # Check the last call to `optimize`, so we catch one during the
            # regular regime
            _, _, kwargs = call.mock_calls[-1]
            self.assertIn("exaggeration", kwargs)
            self.assertEqual(kwargs["exaggeration"], exaggeration)

        # Since optimize needs to return a valid TSNEModel instance and it is
        # impossible to return `self` in a mock, we'll prepare this one ahead
        # of time and use this
        optimize.return_value = DummyTSNE()(self.data)

        # Set value to 1
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.run_button.clicked.emit()  # stop initial run
        self.wait_until_stop_blocking()
        self.widget.controls.exaggeration.setValue(1)
        self.widget.run_button.clicked.emit()  # run with exaggeration 1
        self.wait_until_finished()
        _check_exaggeration(optimize, 1)

        # Reset and clear state
        self.send_signal(self.widget.Inputs.data, None)
        optimize.reset_mock()

        # Change to 3
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.run_button.clicked.emit()  # stop initial run
        self.wait_until_stop_blocking()
        self.widget.controls.exaggeration.setValue(3)
        self.widget.run_button.clicked.emit()  # run with exaggeration 1
        self.wait_until_finished()
        _check_exaggeration(optimize, 3)

    def test_plot_once(self):
        """Test if data is plotted only once but committed on every input change"""
        self.widget.setup_plot = Mock()
        self.widget.commit.deferred = self.widget.commit.now = Mock()

        self.send_signal(self.widget.Inputs.data, self.data)
        # TODO: The base widget immediately calls `setup_plot` and `commit`
        # even though there's nothing to show yet. Unfortunately, fixing this
        # would require changing `OWDataProjectionWidget` in some strange way,
        # so as a temporary fix, we reset the mocks, so they reflect the calls
        # when the result was available.
        self.widget.setup_plot.reset_mock()
        self.widget.commit.deferred.reset_mock()
        self.wait_until_finished()

        self.widget.setup_plot.assert_called_once()
        self.widget.commit.deferred.assert_called_once()

        self.widget.commit.deferred.reset_mock()
        self.send_signal(self.widget.Inputs.data_subset, self.data[::10])
        self.wait_until_stop_blocking()

        self.widget.setup_plot.assert_called_once()
        self.widget.commit.deferred.assert_called_once()

    def test_modified_info_message_behaviour(self):
        """Information messages should be cleared if the data changes or if
        the data is set to None."""
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden by default"
        )

        self.widget.controls.multiscale.setChecked(False)
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden even after toggling "
            "options if no data is on input"
        )

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden after the widget "
            "computes the embedding"
        )

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden when reloading the "
            "same data set and no previous messages were shown"
        )

        self.widget.controls.multiscale.setChecked(True)
        self.assertTrue(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be shown when a setting is "
            "changed, but the embedding is not recomputed"
        )

        self.send_signal(self.widget.Inputs.data, Table("housing"))
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The information message was not cleared on new data"
        )
        # Flip; the info message should now be shown again.
        self.widget.controls.multiscale.setChecked(False)
        assert self.widget.Information.modified.is_shown()  # sanity check

        self.send_signal(self.widget.Inputs.data, None)
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The information message was not cleared on no data"
        )

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden after the widget "
            "computes the embedding"
        )

    def test_invalidation_flow(self):
        # pylint: disable=protected-access
        w = self.widget
        # Setup widget: send data to input with global structure "off", then
        # set global structure "on" (after the embedding is computed)
        w.controls.multiscale.setChecked(False)
        self.send_signal(w.Inputs.data, self.data)

        # By default, t-SNE is smart and disables PCA preprocessing if the
        # number of features is too low. Since we are testing with the iris
        # data set, we want to force t-SNE to use PCA preprocessing.
        w.controls.use_pca_preprocessing.setChecked(True)
        self.widget.run_button.click()

        self.wait_until_finished()
        self.assertFalse(self.widget.Information.modified.is_shown())
        # All the embedding components should be computed
        self.assertIsNotNone(w.preprocessed_data)
        self.assertIsNotNone(w.normalized_data)
        self.assertIsNotNone(w.pca_projection)
        self.assertIsNotNone(w.affinities)
        self.assertIsNotNone(w.tsne_embedding)
        # All the invalidation flags should be set to false
        self.assertFalse(w._invalidated.preprocessed_data)
        self.assertFalse(w._invalidated.normalized_data)
        self.assertFalse(w._invalidated.pca_projection)
        self.assertFalse(w._invalidated.affinities)
        self.assertFalse(w._invalidated.tsne_embedding)

        # Trigger invalidation
        w.controls.multiscale.setChecked(True)
        self.assertTrue(self.widget.Information.modified.is_shown())
        # Setting `multiscale` to true should set the invalidate flags for
        # the affinities and embedding, but not the pca_projection
        self.assertFalse(w._invalidated.preprocessed_data)
        self.assertFalse(w._invalidated.normalized_data)
        self.assertFalse(w._invalidated.pca_projection)
        self.assertTrue(w._invalidated.affinities)
        self.assertTrue(w._invalidated.tsne_embedding)

        # The flags should now be set, but the embedding should still be
        # available when selecting a subset of data and such
        self.assertIsNotNone(w.preprocessed_data)
        self.assertIsNotNone(w.normalized_data)
        self.assertIsNotNone(w.pca_projection)
        self.assertIsNotNone(w.affinities)
        self.assertIsNotNone(w.tsne_embedding)

        # We should still be able to send a data subset to the input and have
        # the points be highlighted
        self.send_signal(w.Inputs.data_subset, self.data[:10])
        self.wait_until_finished()
        subset = [brush.color().name() == "#46befa" for brush in
                  w.graph.scatterplot_item.data["brush"][:10]]
        other = [brush.color().name() == "#000000" for brush in
                 w.graph.scatterplot_item.data["brush"][10:]]
        self.assertTrue(all(subset))
        self.assertTrue(all(other))

        # Clear the data subset
        self.send_signal(w.Inputs.data_subset, None)

        # Run the optimization
        self.widget.run_button.click()
        self.wait_until_finished()
        # All of the inavalidation flags should have been cleared
        self.assertFalse(w._invalidated)

    def test_pca_preprocessing_warning_with_large_number_of_features(self):
        self.assertFalse(
            self.widget.Warning.consider_using_pca_preprocessing.is_shown(),
            "The PCA warning should be hidden by default"
        )

        self.widget.controls.use_pca_preprocessing.setChecked(False)
        self.assertFalse(
            self.widget.Warning.consider_using_pca_preprocessing.is_shown(),
            "The PCA warning should be hidden even after toggling options if "
            "no data is on input"
        )

        # Setup data classes
        x_small = np.random.normal(0, 1, size=(50, 4))
        data_small = Table.from_numpy(Domain.from_numpy(x_small), x_small)
        x_large = np.random.normal(0, 1, size=(50, 250))
        data_large = Table.from_numpy(Domain.from_numpy(x_large), x_large)

        # SMALL data with PCA preprocessing ENABLED
        self.send_signal(self.widget.Inputs.data, data_small)
        self.widget.controls.use_pca_preprocessing.setChecked(True)
        self.widget.run_button.click(), self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Warning.consider_using_pca_preprocessing.is_shown(),
            "The PCA warning should be hidden when PCA preprocessing enabled "
            "when the data has <50 features"
        )

        # SMALL data with PCA preprocessing DISABLED
        self.send_signal(self.widget.Inputs.data, data_small)
        self.widget.controls.use_pca_preprocessing.setChecked(False)
        self.widget.run_button.click(), self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Warning.consider_using_pca_preprocessing.is_shown(),
            "The PCA warning should be hidden with disabled PCA preprocessing "
            "when the data has <50 features"
        )

        # LARGE data with PCA preprocessing ENABLED
        self.send_signal(self.widget.Inputs.data, data_large)
        self.widget.controls.use_pca_preprocessing.setChecked(True)
        self.widget.run_button.click(), self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Warning.consider_using_pca_preprocessing.is_shown(),
            "The PCA warning should be hidden when PCA preprocessing enabled "
            "when has >50 features"
        )

        # LARGE data with PCA preprocessing DISABLED
        self.send_signal(self.widget.Inputs.data, data_large)
        self.widget.controls.use_pca_preprocessing.setChecked(False)
        self.widget.run_button.click(), self.wait_until_stop_blocking()
        self.assertTrue(
            self.widget.Warning.consider_using_pca_preprocessing.is_shown(),
            "The PCA warning should be shown when PCA preprocessing disabled "
            "when the data has >50 features"
        )

        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The PCA warning should be cleared when problematic data is removed"
        )

    def test_distance_matrix_not_symmetric(self):
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_not_symmetric.is_shown())

        self.send_signal(w.Inputs.distances, DistMatrix([[1, 2, 3], [4, 5, 6]]))
        self.assertTrue(w.Error.distance_matrix_not_symmetric.is_shown())

        self.send_signal(w.Inputs.distances, DistMatrix([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        self.assertTrue(w.Error.distance_matrix_not_symmetric.is_shown())

        self.send_signal(w.Inputs.distances, None)
        self.assertFalse(w.Error.distance_matrix_not_symmetric.is_shown())

    def test_matrix_too_small(self):
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, DistMatrix([[1]]))
        self.assertTrue(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, None)
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

    def test_mismatching_distances_and_data_size(self):
        w = self.widget
        self.assertFalse(w.Error.dimension_mismatch.is_shown())

        # Send incompatible combination
        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.send_signal(w.Inputs.data, self.housing)
        self.wait_until_finished()
        self.assertTrue(w.Error.dimension_mismatch.is_shown())

        # Remove offending data
        self.send_signal(w.Inputs.data, None)
        self.wait_until_finished()
        self.assertFalse(w.Error.dimension_mismatch.is_shown())

        # Send incompatible combination
        self.send_signal(w.Inputs.distances, None)
        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.send_signal(w.Inputs.data, self.housing)
        self.wait_until_finished()

        # Remove offending distance matrix
        self.send_signal(w.Inputs.distances, None)
        self.wait_until_finished()
        self.assertFalse(w.Error.dimension_mismatch.is_shown())

        # Clear any data
        self.send_signal(w.Inputs.distances, None)
        self.assertFalse(w.Error.dimension_mismatch.is_shown())

    def test_invalid_distance_matrix_with_valid_data_signal_1(self):
        """Provide valid data table and an invalid distance matrix at the
        same time."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, DistMatrix([[1]]))
        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

    def test_invalid_distance_matrix_with_valid_data_signal_2(self):
        """Provide an invalid distance matrix, wait to finish, then provide a
        valid data table."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, DistMatrix([[1]]))
        self.wait_until_finished()
        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

    def test_invalid_distance_matrix_with_valid_data_signal_3(self):
        """Provide a valid data table, wait to finish, then provide an
        invalid distance matrix."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()
        self.send_signal(w.Inputs.distances, DistMatrix([[1]]))
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

    def test_valid_distance_matrix_with_mismatching_data_signal_1(self):
        """Provide valid distance matrix and a mismatching data table at the
        same time."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.send_signal(w.Inputs.data, self.iris[:5])
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

    def test_valid_distance_matrix_with_mismatching_data_signal_2(self):
        """Provide valid distance matrix, wait to finish, then provide a
        mismatching data table."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.wait_until_finished()
        self.send_signal(w.Inputs.data, self.iris[:5])
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

    def test_invalid_combination_followed_by_valid_combination(self):
        """Provide valid distance matrix and a mismatching data table at the
        same time."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.send_signal(w.Inputs.data, self.iris[:5])
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()
        self.assertIsNotNone(w.graph.scatterplot_item)
        self.assertIsNotNone(self.get_output(w.Outputs.annotated_data))

    def test_invalid_combination_followed_by_valid_distance_only(self):
        """Provide valid distance matrix and a mismatching data table at the
        same time."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.send_signal(w.Inputs.data, self.iris[:5])
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

        self.send_signal(w.Inputs.data, None)
        self.wait_until_finished()
        self.assertIsNotNone(w.graph.scatterplot_item)
        self.assertIsNotNone(self.get_output(w.Outputs.annotated_data))

    def test_invalid_combination_followed_by_valid_data_only(self):
        """Provide valid distance matrix and a mismatching data table at the
        same time."""
        w = self.widget
        self.assertFalse(w.Error.distance_matrix_too_small.is_shown())

        self.send_signal(w.Inputs.distances, DistMatrix([[1]]))
        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()

        self.assertIsNone(w.graph.scatterplot_item)
        self.assertIsNone(self.get_output(w.Outputs.annotated_data))

        self.send_signal(w.Inputs.distances, None)
        self.wait_until_finished()
        self.assertIsNotNone(w.graph.scatterplot_item)
        self.assertIsNotNone(self.get_output(w.Outputs.annotated_data))

    def test_adding_data_table_to_distance_matrix_doesnt_trigger_rerun(self):
        """If the embedding is already constructed, and we just update the data
        signal, then we don't need to recompute the embedding."""
        w = self.widget

        with patch("Orange.widgets.unsupervised.owtsne.TSNERunner.run", wraps=TSNERunner.run) as runner:
            self.send_signal(w.Inputs.distances, Euclidean(self.housing[:150]))
            self.wait_until_finished()

            housing_colors = [
                brush.color().name() for brush in
                w.graph.scatterplot_item.data["brush"]
            ]

            self.send_signal(w.Inputs.data, self.iris)
            self.wait_until_finished()
            iris_colors = [
                brush.color().name() for brush in
                w.graph.scatterplot_item.data["brush"]
            ]

            # Ensure the colors have changed
            self.assertTrue(all(c1 != c2 for c1, c2 in zip(housing_colors, iris_colors)))
            # And that the embedding has not been recomputed
            runner.assert_called_once()

    def test_data_change_doesnt_crash(self):
        w = self.widget

        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()

        self.send_signal(w.Inputs.data, self.housing)
        self.wait_until_finished()

        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()

    def test_distance_change_doesnt_crash(self):
        w = self.widget

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.wait_until_finished()

        self.send_signal(w.Inputs.distances, self.housing_distances)
        self.wait_until_finished()

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.wait_until_finished()

    def test_distances_without_data_axis_0(self):
        w = self.widget
        signal_data = Euclidean(self.data, axis=0)
        signal_data.row_items = None
        self.send_signal(w.Inputs.distances, signal_data)

    def test_distances_without_data_axis_1(self):
        signal_data = Euclidean(self.data, axis=1)
        signal_data.row_items = None
        self.send_signal("Distances", signal_data)

    def test_data_table_with_no_attributes_with_distance_matrix_works(self):
        w = self.widget
        self.send_signal(w.Inputs.distances, self.towns)
        self.wait_until_finished()
        # No errors should be shown
        self.assertEqual(len(w.Error.active), 0)

    def test_controls_are_properly_disabled_with_distance_matrix_1(self):
        """Send both signals first, then disconnect distances."""
        w = self.widget

        disabled_fields = [
            "normalize", "use_pca_preprocessing", "pca_components",
            "initialization_method_idx", "distance_metric_idx",
        ]

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()

        for field in disabled_fields:
            self.assertFalse(getattr(w.controls, field).isEnabled())

        # Remove distance matrix, can use data, so fields should be enabled
        self.send_signal(w.Inputs.distances, None)
        # Ensure PCA checkbox is ticked, to enable slider
        w.controls.use_pca_preprocessing.setChecked(True)
        self.wait_until_finished()

        for field in disabled_fields:
            self.assertTrue(getattr(w.controls, field).isEnabled())

    def test_controls_are_properly_disabled_with_distance_matrix_2(self):
        """Send distances first, disconnect distances, then send data."""
        w = self.widget

        disabled_fields = [
            "normalize", "use_pca_preprocessing", "pca_components",
            "initialization_method_idx", "distance_metric_idx",
        ]

        self.send_signal(w.Inputs.distances, self.iris_distances)
        self.wait_until_finished()

        for field in disabled_fields:
            self.assertFalse(getattr(w.controls, field).isEnabled())

        # Remove distance matrix
        self.send_signal(w.Inputs.distances, None)
        self.wait_until_finished()
        # Send data
        self.send_signal(w.Inputs.data, self.iris)
        # Ensure PCA checkbox is ticked, to enable slider
        w.controls.use_pca_preprocessing.setChecked(True)
        self.wait_until_finished()

        # Should now be enabled
        for field in disabled_fields:
            self.assertTrue(getattr(w.controls, field).isEnabled())

    def test_controls_ignored_by_distance_matrix_retain_values_on_table_signal(self):
        """The controls for `normalize`, `pca_preprocessing`, `metric`, and
        `initialization` are overridden/ignored when using a distance matrix
        signal. However, we want to remember their values when using Data
        table signals."""
        w = self.widget

        # SEND IRIS DATA
        # Set some parameters
        self.send_signal(w.Inputs.data, self.iris)
        w.normalize_cbx.setChecked(False)
        w.pca_preprocessing_cbx.setChecked(True)
        w.pca_component_slider.setValue(3)
        simulate.combobox_activate_index(w.initialization_combo, 0)
        simulate.combobox_activate_index(w.distance_metric_combo, 2)
        w.perplexity_spin.setValue(42)

        # Disconnect data
        self.send_signal(w.Inputs.data, None)

        # SEND IRIS DISTANCES
        self.send_signal(w.Inputs.distances, self.iris_distances)
        # Check that distance-related controls are disabled
        self.assertFalse(w.normalize_cbx.isEnabled())

        self.assertFalse(w.pca_preprocessing_cbx.isEnabled())
        self.assertFalse(w.pca_component_slider.isEnabled())

        self.assertFalse(w.initialization_combo.isEnabled())
        # Only spectral layout is supported when we have distances
        self.assertEqual(w.initialization_combo.currentText(), "Spectral")

        self.assertFalse(w.distance_metric_combo.isEnabled())
        self.assertEqual(w.distance_metric_combo.currentText(), "")

        self.assertTrue(w.perplexity_spin.isEnabled())
        self.assertEqual(w.perplexity_spin.value(), 42)

        # Disconnect distances
        self.send_signal(w.Inputs.distances, None)

        # SEND IRIS DATA
        # The distance-related settings should be restored from when we sent in
        # the data, and not overridden by the settings automatically set by the
        # widget when we passed in the distances signal
        self.send_signal(w.Inputs.data, self.iris)

        # Check that the parameters are restored
        self.assertTrue(w.normalize_cbx.isEnabled())
        self.assertFalse(w.normalize_cbx.isChecked())

        self.assertTrue(w.pca_preprocessing_cbx.isEnabled())
        self.assertTrue(w.pca_preprocessing_cbx.isChecked())
        self.assertTrue(w.pca_component_slider.isEnabled())

        self.assertTrue(w.initialization_combo.isEnabled())
        self.assertTrue(w.initialization_combo.currentText(), "PCA")

        self.assertTrue(w.distance_metric_combo.isEnabled())
        self.assertEqual(w.distance_metric_combo.currentIndex(), 2)

        self.assertTrue(w.perplexity_spin.isEnabled())
        self.assertEqual(w.perplexity_spin.value(), 42)

    def test_controls_are_properly_disabled_with_sparse_matrix(self):
        w = self.widget

        # Normalizing sparse matrix is disabled, since this would require
        # centering
        disabled_fields = ["normalize"]
        # PCA preprocessing and supported distance metrics are enable for sparse
        # matrices
        enabled_fields = [
            "use_pca_preprocessing", "distance_metric_idx", "initialization_method_idx"
        ]

        self.send_signal(w.Inputs.data, self.iris.to_sparse())
        self.wait_until_finished()

        for field in disabled_fields:
            self.assertFalse(getattr(w.controls, field).isEnabled())
        for field in enabled_fields:
            self.assertTrue(getattr(w.controls, field).isEnabled())

        # Send dense table, shoule enable disabled fields
        self.send_signal(w.Inputs.data, self.iris)
        self.wait_until_finished()

        for field in disabled_fields:
            self.assertTrue(getattr(w.controls, field).isEnabled())
        for field in enabled_fields:
            self.assertTrue(getattr(w.controls, field).isEnabled())

    def test_data_containing_nans(self):
        x = np.random.normal(0, 1, size=(150, 50))
        # Randomly sprinkle a few NaNs into the matrix
        num_nans = 20
        x[np.random.randint(0, 150, num_nans), np.random.randint(0, 50, num_nans)] = np.nan

        nan_data = Table.from_numpy(Domain.from_numpy(x), x)

        w = self.widget

        self.send_signal(w.Inputs.data, nan_data)
        self.assertTrue(w.controls.normalize.isChecked())
        self.assertTrue(w.controls.use_pca_preprocessing.isChecked())
        self.widget.run_button.click(), self.wait_until_finished()

        # Disable only normalization
        w.controls.normalize.setChecked(False)
        self.widget.run_button.click(), self.wait_until_finished()

        # Disable only PCA preprocessing
        w.controls.normalize.setChecked(True)
        w.controls.use_pca_preprocessing.setChecked(False)
        self.widget.run_button.click(), self.wait_until_finished()

        # Disable both normalization and PCA preprocessing
        w.controls.normalize.setChecked(False)
        w.controls.use_pca_preprocessing.setChecked(False)
        self.widget.run_button.click(), self.wait_until_finished()


class TestTSNERunner(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = Table("iris")
        cls.distances = Euclidean(cls.data)

    def test_run_with_normalization_and_pca_preprocessing(self):
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)

        task = Task(
            normalize=True,
            use_pca_preprocessing=True,
            data=self.data,
            perplexity=30,
            initialization_method="pca",
            distance_metric="l2",
        )
        task = TSNERunner.run(task, state)

        self.assertEqual(len(state.set_status.mock_calls), 6)
        state.set_status.assert_has_calls([
            call("Preprocessing data..."),
            call("Normalizing data..."),
            call("Computing PCA..."),
            call("Finding nearest neighbors..."),
            call("Preparing initialization..."),
            call("Running optimization..."),
        ])

        self.assertIsInstance(task.normalized_data, Table)
        self.assertIsInstance(task.pca_projection, Table)
        self.assertIsInstance(task.tsne, TSNE)
        self.assertIsInstance(task.tsne_embedding, TSNEModel)

    def test_run_with_normalization(self):
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)

        task = Task(
            normalize=True,
            use_pca_preprocessing=False,
            data=self.data,
            initialization_method="pca",
            distance_metric="l2",
            perplexity=30,
        )
        task = TSNERunner.run(task, state)

        self.assertEqual(len(state.set_status.mock_calls), 5)
        state.set_status.assert_has_calls([
            call("Preprocessing data..."),
            call("Normalizing data..."),
            call("Finding nearest neighbors..."),
            call("Preparing initialization..."),
            call("Running optimization..."),
        ])

        self.assertIsNone(task.pca_projection, Table)

        self.assertIsInstance(task.normalized_data, Table)
        self.assertIsInstance(task.tsne, TSNE)
        self.assertIsInstance(task.tsne_embedding, TSNEModel)

    def test_run_with_pca_preprocessing(self):
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)

        task = Task(
            normalize=False,
            use_pca_preprocessing=True,
            data=self.data,
            initialization_method="pca",
            distance_metric="l2",
            perplexity=30,
        )
        task = TSNERunner.run(task, state)

        self.assertEqual(len(state.set_status.mock_calls), 5)
        state.set_status.assert_has_calls([
            call("Preprocessing data..."),
            call("Computing PCA..."),
            call("Finding nearest neighbors..."),
            call("Preparing initialization..."),
            call("Running optimization..."),
        ])
        state.set_status.assert_has_calls

        self.assertIsNone(task.normalized_data, Table)

        self.assertIsInstance(task.pca_projection, Table)
        self.assertIsInstance(task.tsne, TSNE)
        self.assertIsInstance(task.tsne_embedding, TSNEModel)

    def test_run_do_not_modify_model_inplace(self):
        state = Mock()
        state.is_interruption_requested.return_value = True

        task = Task(
            data=self.data,
            initialization_method="pca",
            distance_metric="l2",
            perplexity=30,
            multiscale=False,
            exaggeration=1,
        )
        # Run through all the steps to prepare the t-SNE object
        task.tsne = prepare_tsne_obj(
            task.data.X.shape[0],
            task.initialization_method,
            task.distance_metric,
            task.perplexity,
            task.multiscale,
            task.exaggeration,
        )
        TSNERunner.compute_normalization(task, state)
        TSNERunner.compute_pca(task, state)
        TSNERunner.compute_initialization(task, state)
        TSNERunner.compute_affinities(task, state)
        # Run the t-SNE iteration once to create the object
        TSNERunner.compute_tsne(task, state)

        # Make sure that running t-SNE for another iteration returns a new object
        tsne_obj_before = task.tsne_embedding
        state.reset_mock()
        TSNERunner.compute_tsne(task, state)
        tsne_obj_after = task.tsne_embedding

        state.set_partial_result.assert_called_once()
        self.assertIsNot(tsne_obj_before, tsne_obj_after)

    def test_run_with_distance_matrix(self):
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)

        task = Task(
            normalize=False,
            use_pca_preprocessing=False,
            distance_matrix=self.distances,
            perplexity=30,
            initialization_method="spectral",
            distance_metric="precomputed",
        )
        task = TSNERunner.run(task, state)

        self.assertEqual(len(state.set_status.mock_calls), 3)
        state.set_status.assert_has_calls([
            call("Finding nearest neighbors..."),
            call("Preparing initialization..."),
            call("Running optimization..."),
        ])

        self.assertIsNone(task.normalized_data)
        self.assertIsNone(task.pca_projection)
        self.assertIsInstance(task.initialization, np.ndarray)
        self.assertIsInstance(task.tsne, TSNE)
        self.assertIsInstance(task.tsne_embedding, TSNEModel)

    def test_task_validation(self):
        # distance matrix with no data table
        Task(
            normalize=True,
            use_pca_preprocessing=True,
            distance_matrix=self.distances,
            perplexity=30,
            initialization_method="spectral",
            distance_metric="precomputed",
        ).validate()

        # both distance matrix and data table are provided
        Task(
            normalize=True,
            use_pca_preprocessing=True,
            data=self.data,
            distance_matrix=self.distances,
            perplexity=30,
            initialization_method="spectral",
            distance_metric="precomputed",
        ).validate()

        # data table with no distance matrix
        Task(
            normalize=True,
            use_pca_preprocessing=True,
            data=self.data,
            perplexity=30,
            initialization_method="pca",
            distance_metric="cosine",
        ).validate()

        # distance_metric="precomputed" with no distance matrix
        with self.assertRaises(Task.ValidationError):
            Task(
                normalize=True,
                use_pca_preprocessing=True,
                data=self.data,
                perplexity=30,
                initialization_method="spectral",
                distance_metric="precomputed",
            ).validate()

        # initialization_method="pca" with distance matrix
        with self.assertRaises(Task.ValidationError):
            Task(
                normalize=True,
                use_pca_preprocessing=True,
                data=self.data,
                distance_matrix=self.distances,
                perplexity=30,
                initialization_method="pca",
                distance_metric="precomputed",
            ).validate()

        # distance_metric="l2" with distance matrix
        with self.assertRaises(Task.ValidationError):
            Task(
                normalize=True,
                use_pca_preprocessing=True,
                data=self.data,
                distance_matrix=self.distances,
                perplexity=30,
                initialization_method="spectral",
                distance_metric="l2",
            ).validate()

    def test_run_with_distance_matrix_ignores_preprocessing(self):
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)

        task = Task(
            normalize=True,
            use_pca_preprocessing=True,
            distance_matrix=self.distances,
            perplexity=30,
            initialization_method="spectral",
            distance_metric="precomputed",
        )
        task = TSNERunner.run(task, state)

        self.assertEqual(len(state.set_status.mock_calls), 3)
        state.set_status.assert_has_calls([
            call("Finding nearest neighbors..."),
            call("Preparing initialization..."),
            call("Running optimization..."),
        ])

        self.assertIsNone(task.normalized_data)
        self.assertIsNone(task.pca_projection)
        self.assertIsInstance(task.initialization, np.ndarray)
        self.assertIsInstance(task.tsne, TSNE)
        self.assertIsInstance(task.tsne_embedding, TSNEModel)

    def test_run_with_sparse_matrix_ignores_normalization(self):
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)

        task = Task(
            normalize=False,
            use_pca_preprocessing=True,
            data=self.data.to_sparse(),
            perplexity=30,
            initialization_method="spectral",
            distance_metric="cosine",
        )
        task = TSNERunner.run(task, state)
        self.assertEqual(len(state.set_status.mock_calls), 5)
        state.set_status.assert_has_calls([
            call("Preprocessing data..."),
            call("Computing PCA..."),
            call("Finding nearest neighbors..."),
            call("Preparing initialization..."),
            call("Running optimization..."),
        ])

        self.assertIsNone(task.normalized_data)
        self.assertIsInstance(task.pca_projection, Table)
        self.assertIsInstance(task.initialization, np.ndarray)
        self.assertIsInstance(task.tsne, TSNE)
        self.assertIsInstance(task.tsne_embedding, TSNEModel)


if __name__ == "__main__":
    unittest.main()



from typing import Optional

from Orange.data import Table
from Orange.widgets import gui
from Orange.widgets.report.report import describe_data
from Orange.widgets.utils.sql import check_sql_input
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from Orange.widgets.utils.concurrents import TaskState, ConcurrentWidgetMixin

from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.data.data.owtransform." + key)
class TransformRunner:
    @staticmethod
    def run(
            data: Table,
            template_data: Table,
            state: TaskState
    ) -> Optional[Table]:
        if data is None or template_data is None:
            return None

class OWTransform(OWWidget, ConcurrentWidgetMixin):
    name = __("name")
    description = __("desc")
    category = i18n.t("widget.data.data.package_label_transform")
    icon = "icons/Transform.svg"
    priority = 1230
    keywords = "apply domain, transform"

    class Inputs:
        data = Input("Data", Table, default=True, label=i18n.t("widget.data.data.common.data"))
        template_data = Input("Template Data", Table, label=i18n.t("widget.data.data.common.template_data"))

    class Outputs:
        transformed_data = Output("Transformed Data", Table, label=i18n.t("widget.data.data.common.transformed_data"))

    class Error(OWWidget.Error):
        error = Msg(__("msg_transform_error"))

    resizing_enabled = False
    want_main_area = False
    buttons_area_orientation = None

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)
        self.data = None  # type: Optional[Table]
        self.template_data = None  # type: Optional[Table]
        self.transformed_info = describe_data(None)  # type: OrderedDict

        box = gui.widgetBox(self.controlArea, True)
        gui.label(
            box, self, __("text.tip_text").strip(), box=True)

    @Inputs.data
    @check_sql_input
    def set_data(self, data):
        self.data = data

    @Inputs.template_data
    @check_sql_input
    def set_template_data(self, data):
        self.template_data = data

    def handleNewSignals(self):
        self.apply()

    def apply(self):
        self.clear_messages()
        self.cancel()
        self.start(TransformRunner.run, self.data, self.template_data)

    def send_report(self):
        if self.data:
            self.report_data(__("report.data"), self.data)
        if self.template_data:
            self.report_domain(__("report.template_data"), self.template_data.domain)
        if self.transformed_info:
            self.report_items(__("report.transformed_data"), self.transformed_info)

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Optional[Table]):
        self.transformed_info = describe_data(result)
        self.Outputs.transformed_data.send(result)

    def on_exception(self, ex):
        self.Error.error(ex)
        self.Outputs.transformed_data.send(None)

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()



if __name__ == "__main__":  # pragma: no cover
    from Orange.preprocess import Discretize

    table = Table("iris")
    WidgetPreview(OWTransform).run(
        set_data=table, set_template_data=Discretize()(table))



import i18n
import os

localedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'locale', 'orange')

if localedir not in i18n.load_path:
    i18n.load_path.append(localedir)

# import os
# import configparser
# def read_ini_value():
#     try:
#         # 获取用户个人目录
#         user_profile = os.getenv('USERPROFILE') or os.getenv('HOME')
#         if user_profile is None:
#             raise EnvironmentError("环境变量 USERPROFILE 或 HOME 未设置")
        
#         # 构造完整路径（Windows路径和Linux/macOS路径不同）
#         if os.name == 'nt':  # Windows
#             file_path = os.path.join(user_profile, 'AppData', 'Roaming', 'biolab.si', 'language.ini')
#         else:  # macOS 和 Linux
#             file_path = os.path.join(user_profile, '.config', 'biolab.si', 'BlueWhale.ini')

#         # 检查文件是否存在
#         if not os.path.isfile(file_path):
#             raise FileNotFoundError(f"文件未找到: {file_path}")

#         # 创建一个ConfigParser对象
#         config = configparser.ConfigParser()
#         config.read(file_path)

#         # 获取并打印指定键的值
#         value = config.get('application-style', 'languages')
#         return value

#     except (EnvironmentError, FileNotFoundError) as e:
#         print(f"路径错误: {e}")
#     except (configparser.NoSectionError, configparser.NoOptionError) as e:
#         print(f"配置文件错误: {e}")
#     except Exception as e:
#         print(f"读取文件时发生错误: {e}")

#     # 返回默认值
#     return 'en'


# # 主程序调用
# default_value = read_ini_value()
# i18n.set('locale', default_value)
# print(default_value)


import unittest
import numpy as np

from Orange.data import Table, Domain, StringVariable
from Orange.widgets.model.owpls import OWPLS
from Orange.widgets.tests.base import WidgetTest, WidgetLearnerTestMixin, \
    ParameterMapping


class TestOWPLS(WidgetTest, WidgetLearnerTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._data = Table("housing")
        cls._data = cls._data.add_column(StringVariable("Foo"),
                                         ["Bar"] * len(cls._data),
                                         to_metas=True)
        class_vars = [cls._data.domain.class_var,
                      cls._data.domain.attributes[0]]
        domain = Domain(cls._data.domain.attributes[1:], class_vars,
                        cls._data.domain.metas)
        cls._data_multi_target = cls._data.transform(domain)

    def setUp(self):
        self.widget = self.create_widget(OWPLS,
                                         stored_settings={"auto_apply": False})
        self.init()
        self.parameters = [
            ParameterMapping('max_iter', self.widget.controls.max_iter),
            ParameterMapping('n_components', self.widget.controls.n_components)
        ]

    def test_output_coefsdata(self):
        self.send_signal(self.widget.Inputs.data, self._data)
        coefsdata = self.get_output(self.widget.Outputs.coefsdata)
        self.assertEqual(coefsdata.name, "Coefficients and Loadings")
        self.assertEqual(coefsdata.X.shape, (15, 4))
        self.assertEqual(coefsdata.Y.shape, (15, 0))
        self.assertEqual(coefsdata.metas.shape, (15, 2))

        self.assertEqual(["coef (MEDV)", "coef/X_sd (MEDV)", "w*c 1", "w*c 2"],
                         [v.name for v in coefsdata.domain.attributes])
        self.assertEqual(["Variable name", "Variable role"],
                         [v.name for v in coefsdata.domain.metas])
        metas = [v.name for v in self._data.domain.variables] + ["intercept"]
        self.assertTrue((coefsdata.metas[:, 0] == metas).all())
        self.assertTrue((coefsdata.metas[:-2, 1] == 0).all())
        self.assertTrue((coefsdata.metas[-2, 1] == 1))
        self.assertTrue(np.isnan(coefsdata.metas[-1, 1]))
        self.assertAlmostEqual(coefsdata.X[0, 2], 0.237, 3)
        self.assertAlmostEqual(coefsdata.X[13, 2], -0.304, 3)
        self.assertAlmostEqual(coefsdata.X[-1, 0], 22.5, 1)
        self.assertTrue(np.isnan(coefsdata.X[-1, 2:]).all())

    def test_output_coefsdata_multi_target(self):
        self.send_signal(self.widget.Inputs.data, self._data_multi_target)
        coefsdata = self.get_output(self.widget.Outputs.coefsdata)
        self.assertEqual(coefsdata.name, "Coefficients and Loadings")
        self.assertEqual(coefsdata.X.shape, (15, 6))
        self.assertEqual(coefsdata.Y.shape, (15, 0))
        self.assertEqual(coefsdata.metas.shape, (15, 2))

        attr_names = ["coef (MEDV)", "coef (CRIM)", "coef/X_sd (MEDV)",
                      "coef/X_sd (CRIM)", "w*c 1", "w*c 2"]
        self.assertEqual(attr_names,
                         [v.name for v in coefsdata.domain.attributes])
        self.assertEqual(["Variable name", "Variable role"],
                         [v.name for v in coefsdata.domain.metas])
        metas = [v.name for v in self._data_multi_target.domain.variables]
        metas += ["intercept"]
        self.assertTrue((coefsdata.metas[:, 0] == metas).all())
        self.assertTrue((coefsdata.metas[:-3, 1] == 0).all())
        self.assertTrue((coefsdata.metas[-2:-1, 1] == 1).all())
        self.assertTrue(np.isnan(coefsdata.metas[-1, 1]))
        self.assertAlmostEqual(coefsdata.X[0, 4], -0.198, 3)
        self.assertAlmostEqual(coefsdata.X[12, 4], -0.288, 3)
        self.assertAlmostEqual(coefsdata.X[13, 4], 0.243, 3)
        self.assertAlmostEqual(coefsdata.X[-1, 0], 22.5, 1)
        self.assertAlmostEqual(coefsdata.X[-1, 1], 3.6, 1)
        self.assertTrue(np.isnan(coefsdata.X[-1, 4:]).all())

    def test_output_data(self):
        self.send_signal(self.widget.Inputs.data, self._data)
        output = self.get_output(self.widget.Outputs.data)
        self.assertEqual(output.X.shape, (506, 13))
        self.assertEqual(output.Y.shape, (506,))
        self.assertEqual(output.metas.shape, (506, 8))
        self.assertEqual([v.name for v in self._data.domain.variables],
                         [v.name for v in output.domain.variables])
        metas = ["PLS T1", "PLS T2", "PLS U1", "PLS U2",
                 "Sample Quantiles (MEDV)", "Theoretical Quantiles (MEDV)",
                 "DModX"]
        self.assertEqual([v.name for v in self._data.domain.metas] + metas,
                         [v.name for v in output.domain.metas])

    def test_output_data_multi_target(self):
        self.send_signal(self.widget.Inputs.data, self._data_multi_target)
        output = self.get_output(self.widget.Outputs.data)
        self.assertEqual(output.X.shape, (506, 12))
        self.assertEqual(output.Y.shape, (506, 2))
        self.assertEqual(output.metas.shape, (506, 10))
        orig_domain = self._data_multi_target.domain
        self.assertEqual([v.name for v in orig_domain.variables],
                         [v.name for v in output.domain.variables])
        metas = ["PLS T1", "PLS T2", "PLS U1", "PLS U2",
                 "Sample Quantiles (MEDV)", "Theoretical Quantiles (MEDV)",
                 "Sample Quantiles (CRIM)", "Theoretical Quantiles (CRIM)",
                 "DModX"]
        self.assertEqual([v.name for v in orig_domain.metas] + metas,
                         [v.name for v in output.domain.metas])

    def test_output_components(self):
        self.send_signal(self.widget.Inputs.data, self._data)
        components = self.get_output(self.widget.Outputs.components)
        self.assertEqual(components.X.shape, (2, 13))
        self.assertEqual(components.Y.shape, (2,))
        self.assertEqual(components.metas.shape, (2, 1))

    def test_output_components_multi_target(self):
        self.send_signal(self.widget.Inputs.data, self._data_multi_target)
        components = self.get_output(self.widget.Outputs.components)
        self.assertEqual(components.X.shape, (2, 12))
        self.assertEqual(components.Y.shape, (2, 2))
        self.assertEqual(components.metas.shape, (2, 1))

    def test_missing_target(self):
        data = self._data[:5].copy()
        data.Y[[0, 4]] = np.nan
        self.send_signal(self.widget.Inputs.data, data)
        output = self.get_output(self.widget.Outputs.data)
        self.assertFalse(np.isnan(output.metas[:, 1:3].astype(float)).any())
        self.assertTrue(np.isnan(output.metas[0, 3:4].astype(float)).all())
        self.assertTrue(np.isnan(output.metas[4, 3:5].astype(float)).all())
        self.assertFalse(np.isnan(output.metas[1:4, 3:5].astype(float)).any())

        with data.unlocked(data.Y):
            data.Y[:] = np.nan
        self.send_signal(self.widget.Inputs.data, data)
        self.assertIsNone(self.get_output(self.widget.Outputs.data))

    def test_scale(self):
        self.widget.auto_apply = True
        self.send_signal(self.widget.Inputs.data, self._data)
        output1 = self.get_output(self.widget.Outputs.data)
        self.widget.controls.scale.setChecked(False)
        output2 = self.get_output(self.widget.Outputs.data)
        self.assertTrue(abs(output1.metas[0, 1] - output2.metas[0, 1]) > 100)


if __name__ == "__main__":
    unittest.main()



from itertools import count
import time
import os
import unittest
from unittest.mock import Mock, patch
import warnings

import numpy as np
import scipy.sparse as sp

from Orange.util import export_globals, flatten, deprecated, try_, deepgetattr, \
    OrangeDeprecationWarning, nan_eq, nan_hash_stand, Registry, namegen
from Orange.data.util import vstack, hstack, array_equal
from Orange.statistics.util import stats
from Orange.tests.test_statistics import dense_sparse
from Orange.util import wrap_callback, get_entry_point, allot

SOMETHING = 0xf00babe


class TestUtil(unittest.TestCase):
    def test_get_entry_point(self):
        # pylint: disable=import-outside-toplevel
        from Orange.canvas.__main__ import main as real_main
        main = get_entry_point("Orange3", "gui_scripts", "orange-canvas")
        self.assertIs(main, real_main)

    def test_export_globals(self):
        self.assertEqual(sorted(export_globals(globals(), __name__)),
                         ['SOMETHING', 'TestAllot', 'TestUtil'])

    def test_flatten(self):
        self.assertEqual(list(flatten([[1, 2], [3]])), [1, 2, 3])

    def test_deprecated(self):
        @deprecated
        def identity(x):
            return x

        with self.assertWarns(DeprecationWarning) as cm:
            x = identity(10)
        self.assertEqual(x, 10)
        self.assertTrue('deprecated' in cm.warning.args[0])
        self.assertTrue('identity' in cm.warning.args[0])

    def test_try_(self):
        self.assertTrue(try_(lambda: np.ones(3).any()))
        self.assertFalse(try_(lambda: 1 / 0))
        self.assertEqual(try_(len, default=SOMETHING), SOMETHING)

    def test_reprable(self):
        # pylint: disable=import-outside-toplevel
        from Orange.data import ContinuousVariable
        from Orange.preprocess.impute import ReplaceUnknownsRandom
        from Orange.statistics.distribution import Continuous
        from Orange.classification import LogisticRegressionLearner

        var = ContinuousVariable('x')
        transform = ReplaceUnknownsRandom(var, Continuous(1, var))

        self.assertEqual(repr(transform).replace('\n', '').replace(' ', ''),
                         "ReplaceUnknownsRandom("
                         "variable=ContinuousVariable(name='x',number_of_decimals=3),"
                         "distribution=Continuous([[0.],[0.]]))")

        # GH 2275
        logit = LogisticRegressionLearner()
        for _ in range(2):
            self.assertEqual(repr(logit), 'LogisticRegressionLearner()')

    def test_deepgetattr(self):
        class a:  # pylint: disable=invalid-name
            l = []
        self.assertTrue(deepgetattr(a, 'l.__len__.__call__'), a.l.__len__.__call__)
        self.assertTrue(deepgetattr(a, 'l.__nx__.__x__', 42), 42)
        self.assertRaises(AttributeError, lambda: deepgetattr(a, 'l.__nx__.__x__'))

    def test_nan_eq(self):
        self.assertTrue(nan_eq(float("nan"), float("nan")))
        self.assertTrue(nan_eq(1, 1.0))
        self.assertFalse(nan_eq(float("nan"), 1))
        self.assertFalse(nan_eq(1, float("nan")))
        self.assertFalse(nan_eq(float("inf"), float("nan")))
        self.assertFalse(nan_eq(float("nan"), float("inf")))
        self.assertFalse(nan_eq(1, 2))
        self.assertFalse(nan_eq(None, 2))
        self.assertFalse(nan_eq(2, None))

    def test_nan_hash_stand(self):
        self.assertEqual(hash(nan_hash_stand(float("nan"))),
                         hash(nan_hash_stand(float("nan"))))

    def test_vstack(self):
        numpy = np.array([[1., 2.], [3., 4.]])
        csr = sp.csr_matrix(numpy)
        csc = sp.csc_matrix(numpy)

        self.assertCorrectArrayType(
            vstack([numpy, numpy]),
            shape=(4, 2), sparsity="dense")
        self.assertCorrectArrayType(
            vstack([csr, numpy]),
            shape=(4, 2), sparsity="sparse")
        self.assertCorrectArrayType(
            vstack([numpy, csc]),
            shape=(4, 2), sparsity="sparse")
        self.assertCorrectArrayType(
            vstack([csc, csr]),
            shape=(4, 2), sparsity="sparse")

    def test_hstack(self):
        numpy = np.array([[1., 2.], [3., 4.]])
        csr = sp.csr_matrix(numpy)
        csc = sp.csc_matrix(numpy)

        self.assertCorrectArrayType(
            hstack([numpy, numpy]),
            shape=(2, 4), sparsity="dense")
        self.assertCorrectArrayType(
            hstack([csr, numpy]),
            shape=(2, 4), sparsity="sparse")
        self.assertCorrectArrayType(
            hstack([numpy, csc]),
            shape=(2, 4), sparsity="sparse")
        self.assertCorrectArrayType(
            hstack([csc, csr]),
            shape=(2, 4), sparsity="sparse")

    def assertCorrectArrayType(self, array, shape, sparsity):
        self.assertEqual(array.shape, shape)
        self.assertEqual(["dense", "sparse"][sp.issparse(array)], sparsity)

    @unittest.skipUnless(os.environ.get('ORANGE_DEPRECATIONS_ERROR'),
                         'ORANGE_DEPRECATIONS_ERROR not set')
    def test_raise_deprecations(self):
        with self.assertRaises(OrangeDeprecationWarning):
            warnings.warn('foo', OrangeDeprecationWarning)

    def test_stats_sparse(self):
        # pylint: disable=import-outside-toplevel
        from Orange.data import Table
        data = Table("iris")
        sparse_x = sp.csr_matrix(data.X)
        self.assertTrue(stats(data.X).all() == stats(sparse_x).all())

    @dense_sparse
    def test_array_equal(self, array):
        a1 = array([[0., 2.], [3., np.nan]])
        a2 = array([[0., 2.], [3., np.nan]])
        self.assertTrue(array_equal(a1, a2))

        a3 = np.array([[0., 2.], [3., np.nan]])
        self.assertTrue(array_equal(a1, a3))
        self.assertTrue(array_equal(a3, a1))

    @dense_sparse
    def test_array_not_equal(self, array):
        a1 = array([[0., 2.], [3., np.nan]])
        a2 = array([[0., 2.], [4., np.nan]])
        self.assertFalse(array_equal(a1, a2))

        a3 = array([[0., 2.], [3., np.nan], [4., 5.]])
        self.assertFalse(array_equal(a1, a3))

    def test_csc_array_equal(self):
        a1 = sp.csc_matrix(([1, 4, 5], ([0, 0, 1], [0, 2, 2])), shape=(2, 3))
        a2 = sp.csc_matrix(([5, 1, 4], ([1, 0, 0], [2, 0, 2])), shape=(2, 3))
        with warnings.catch_warnings():
            # this is just inefficiency in tests, not the tested code
            warnings.filterwarnings("ignore", ".*", sp.SparseEfficiencyWarning)
            a2[0, 1] = 0  # explicitly setting to 0
        self.assertTrue(array_equal(a1, a2))

    def test_csc_scr_equal(self):
        a1 = sp.csc_matrix(([1, 4, 5], ([0, 0, 1], [0, 2, 2])), shape=(2, 3))
        a2 = sp.csr_matrix(([5, 1, 4], ([1, 0, 0], [2, 0, 2])), shape=(2, 3))
        self.assertTrue(array_equal(a1, a2))

        a1 = sp.csc_matrix(([1, 4, 5], ([0, 0, 1], [0, 2, 2])), shape=(2, 3))
        a2 = sp.csr_matrix(([1, 4, 5], ([0, 0, 1], [0, 2, 2])), shape=(2, 3))
        self.assertTrue(array_equal(a1, a2))

    def test_csc_unordered_array_equal(self):
        a1 = sp.csc_matrix(([1, 4, 5], [0, 0, 1], [0, 1, 1, 3]), shape=(2, 3))
        a2 = sp.csc_matrix(([1, 5, 4], [0, 1, 0], [0, 1, 1, 3]), shape=(2, 3))
        self.assertTrue(array_equal(a1, a2))

    def test_wrap_callback(self):
        def func(i):
            return i

        f = wrap_callback(func, start=0, end=0.8)
        self.assertEqual(f(0), 0)
        self.assertEqual(round(f(0.1), 2), 0.08)
        self.assertEqual(f(1), 0.8)

        f = wrap_callback(func, start=0.1, end=0.8)
        self.assertEqual(f(0), 0.1)
        self.assertEqual(f(0.1), 0.17)
        self.assertEqual(f(1), 0.8)

    def test_registry(self):
        # pylint: disable=invalid-name, unused-variable
        class A(metaclass=Registry):
            pass

        class BBB(A):
            pass

        class CC(A):
            pass

        class DDDD(BBB):
            pass

        self.assertEqual(set(A), {"BBB", "CC", "DDDD"})
        self.assertEqual(str(A), "A({BBB, CC, DDDD})")

        class D(metaclass=Registry):
            pass

        self.assertEqual(set(A), {"BBB", "CC", "DDDD"})
        self.assertEqual(str(A), "A({BBB, CC, DDDD})")
        self.assertEqual(set(D), set())
        self.assertEqual(str(D), "D({})")

        class E(D):
            pass

        self.assertEqual(set(A), {"BBB", "CC", "DDDD"})
        self.assertEqual(str(A), "A({BBB, CC, DDDD})")
        self.assertEqual(set(D), set("E"))
        self.assertEqual(str(D), "D({E})")

    def test_namegen(self):
        self.assertEqual([name for name, _ in zip(namegen(), range(3))],
                         ["_0", "_1", "_2"])

        self.assertEqual([name for name, _ in zip(namegen("foo "), range(3))],
                         ["foo 0", "foo 1", "foo 2"])

        self.assertEqual([name for name, _ in zip(namegen("foo ", 2, spec_count=count), range(3))],
                         ["foo 2", "foo 3", "foo 4"])


class TestAllot(unittest.TestCase):
    # names of functions within tests don't matter, pylint: disable=invalid-name

    def setUp(self):
        # patch the object to user perf_counter, which will include the time
        # when tests `sleep`
        patcher = patch.object(allot, "_allot__timer", new=time.perf_counter)
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_duration(self):
        @allot
        def f(x, y):
            self.assertEqual(x, 5)
            self.assertEqual(y, 6)
            time.sleep(0.2)

        f(5, y=6)
        self.assertGreaterEqual(f.last_call_duration, 0.2)

        class A:
            def __init__(self):
                self.x = self.y = 0

            @allot
            def f(self, x, y):
                self.x = x
                self.y = y
                time.sleep(0.2)

        a = A()
        a.f(5, y=6)
        self.assertEqual(a.x, 5)
        self.assertEqual(a.y, 6)
        self.assertGreaterEqual(f.last_call_duration, 0.2)

    def test_skipping(self):
        uf = Mock()

        @allot(0.5)
        def f(x, y):
            uf()
            self.assertEqual(x, 5)
            self.assertEqual(y, 6)
            time.sleep(0.2)

        f(5, y=6)
        uf.assert_called_once()
        uf.reset_mock()
        self.assertGreaterEqual(f.last_call_duration, 0.2)
        f(5, y=6)
        uf.assert_not_called()
        self.assertGreaterEqual(f.last_call_duration, 0.2)
        time.sleep(0.35)
        f(5, y=6)
        uf.assert_called_once()
        self.assertGreaterEqual(f.last_call_duration, 0.2)

        class A:
            def __init__(self):
                self.x = self.y = 0

            @allot(0.5)
            def f(self, x, y):
                self.x = x
                self.y = y
                time.sleep(0.2)

        a = A()
        a2 = A()
        a.f(5, y=6)
        self.assertEqual(a.x, 5)
        self.assertEqual(a.y, 6)
        self.assertGreaterEqual(f.last_call_duration, 0.2)

        a.f(7, y=8)
        self.assertEqual(a.x, 5)  # no call, `a` is unchanged
        self.assertGreaterEqual(f.last_call_duration, 0.2)

        time.sleep(0.35)
        a.f(9, y=10)
        a2.f(11, y=12)
        self.assertEqual(a.x, 9)
        self.assertGreaterEqual(f.last_call_duration, 0.2)
        # a2.f is not being skipped because of a.f - bound methods for different
        # instance are separate
        self.assertEqual(a2.x, 11)

        # forced calls work
        a.f.call(11, 12)
        self.assertEqual(a.x, 11)

        # forced calls block later non-forced calls
        a = A()
        a.f.call(13, y=14)
        self.assertEqual(a.x, 13)
        a.f(15, y=16)
        self.assertEqual(a.x, 13)

    def test_overflow(self):
        uf = Mock()
        of = Mock(return_value=13)

        @allot(0.1, overflow=of)
        def f(*_, **__):
            uf()
            time.sleep(0.1)
            return 42

        self.assertEqual(f(5, y=6), 42)
        uf.assert_called()
        uf.reset_mock()
        of.assert_not_called()

        self.assertEqual(f(7, y=8), 13)
        uf.assert_not_called()
        of.assert_called_with(7, y=8)

    def test_assert_no_result(self):
        # pylint: disable=function-redefined
        @allot(0.1)
        def f():
            return 42

        self.assertRaises(AssertionError, f)

        @allot(0.05, overflow=lambda x: 3 * x)
        def f(x):
            time.sleep(0.6)
            return 2 * x

        self.assertEqual(f(3), 6)
        self.assertEqual(f(3), 9)

        @allot
        def f():
            return 42

        self.assertEqual(f(), 42)

    def test_assertion(self):
        self.assertRaises(AssertionError, allot, "foo")
        self.assertRaises(AssertionError, allot, 0.0)
        self.assertRaises(AssertionError, allot, -1.0)

    def test_getter(self):
        class A:
            @allot
            def f(self):
                pass

        # getter for class must not do anything with the method
        self.assertIs(A.f, A.__dict__["f"])


if __name__ == "__main__":
    unittest.main()



from Orange.base import Learner, Model, SklLearner, SklModel
from Orange.i18n_config import *
def __(key):
    return i18n.t("orange." + key)

__all__ = ["LearnerClassification", "ModelClassification",
           "SklModelClassification", "SklLearnerClassification"]


class LearnerClassification(Learner):
    learner_adequacy_err_msg = __("tip.need_discrete_class_variable")

    def incompatibility_reason(self, domain):
        reason = None
        if len(domain.class_vars) > 1 and not self.supports_multiclass:
            reason = "Too many target variables."
        elif not domain.has_discrete_class:
            reason = "Categorical class variable expected."
        return reason


class ModelClassification(Model):
    def predict_proba(self, data):
        return self(data, ret=Model.Probs)


class SklModelClassification(SklModel, ModelClassification):
    pass


class SklLearnerClassification(SklLearner, LearnerClassification):
    __returns__ = SklModelClassification



from unittest import TestCase

from packaging.version import Version


class TestPkgResources(TestCase):
    def test_parse_version(self):
        self.assertGreater(Version('3.4.1'), Version('3.4.0'))
        self.assertGreater(Version('3.4.1'), Version('3.4.dev'))
        self.assertGreater(Version('3.4.1'), Version('3.4.1.dev'))
        self.assertLess(Version('3.4.1'), Version('3.4.2.dev'))



from typing import NamedTuple, Dict, Type, Optional

from AnyQt.QtWidgets import QButtonGroup, QRadioButton
from AnyQt.QtCore import Qt
from scipy.sparse import issparse
import bottleneck as bn

import Orange.data
import Orange.misc
from Orange import distance
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.utils.concurrents import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.sql import check_sql_input
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Msg, Input, Output

from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.unsupervised.unsupervised.owdistances." + key)


Euclidean, EuclideanNormalized, Manhattan, ManhattanNormalized, Cosine, \
    Mahalanobis, Hamming, \
    Pearson, PearsonAbsolute, Spearman, SpearmanAbsolute, Jaccard = range(12)


class MetricDef(NamedTuple):
    id: int  # pylint: disable=invalid-name
    name: str
    tooltip: str
    metric: Type[distance.Distance]
    normalize: bool = False


MetricDefs: Dict[int, MetricDef] = {
    metric.id: metric for metric in (
        MetricDef(EuclideanNormalized, __("euclidean_normalized"),
                  "Square root of summed difference between normalized values",
                  distance.Euclidean, normalize=True),
        MetricDef(Euclidean, __("euclidean"),
                  "Square root of summed difference between values",
                  distance.Euclidean),
        MetricDef(ManhattanNormalized, __("manhattan_normalized"),
                  "Sum of absolute differences between normalized values",
                  distance.Manhattan, normalize=True),
        MetricDef(Manhattan, __("manhattan"),
                  "Sum of absolute differences between values",
                  distance.Manhattan),
        MetricDef(Mahalanobis, __("mahalanobis"),
                  "Mahalanobis distance",
                  distance.Mahalanobis),
        MetricDef(Hamming, __("hamming"),"Hamming distance",
                  distance.Hamming),
        MetricDef(Cosine,  __("cosine"),"Cosine distance",
                  distance.Cosine),
        MetricDef(Pearson, __("pearson"),
                  "Pearson correlation; distance = 1 - ρ/2",
                  distance.PearsonR),
        MetricDef(PearsonAbsolute, __("pearson_absolute"),
                  "Absolute value of Pearson correlation; distance = 1 - |ρ|",
                  distance.PearsonRAbsolute),
        MetricDef(Spearman, __("spearman"),
                  "Spearman correlation; distance = 1 - ρ/2",
                  distance.SpearmanR),
        MetricDef(SpearmanAbsolute, __("spearman_absolute"),
                  "Absolute value of Pearson correlation; distance = 1 - |ρ|",
                  distance.SpearmanRAbsolute),
        MetricDef(Jaccard,  __("jaccard"),"Jaccard distance",
                  distance.Jaccard)
    )
}


class InterruptException(Exception):
    pass


class DistanceRunner:
    @staticmethod
    def run(data: Orange.data.Table, metric: distance, normalized_dist: bool,
            axis: int, state: TaskState) -> Optional[Orange.misc.DistMatrix]:
        if data is None:
            return None

        def callback(i: float) -> bool:
            state.set_progress_value(i)
            if state.is_interruption_requested():
                raise InterruptException

        state.set_status(__("set_status_calculating"))
        kwargs = {"axis": 1 - axis, "impute": True, "callback": callback}
        if metric.supports_normalization and normalized_dist:
            kwargs["normalize"] = True
        return metric(data, **kwargs)


class OWDistances(OWWidget, ConcurrentWidgetMixin):
    name = __("name")
    description = __("desc")
    icon = "icons/Distance.svg"
    keywords = "distances"

    class Inputs:
        data = Input('Data', Orange.data.Table, label=i18n.t("widget.unsupervised.unsupervised.common.data"))

    class Outputs:
        distances = Output('Distances', Orange.misc.DistMatrix, dynamic=False,
                           label=i18n.t("common.general.distance"))

    settings_version = 4

    axis: int = Setting(0)
    metric_id: int = Setting(EuclideanNormalized)
    autocommit: bool = Setting(True)

    want_main_area = False
    resizing_enabled = False

    class Error(OWWidget.Error):
        no_continuous_features = Msg(__("msg.no_numeric"))
        no_binary_features = Msg(__("msg.no_binary"))
        dense_metric_sparse_data = Msg(__("msg.dense_data"))
        distances_memory_error = Msg(i18n.t("common.software.not_enough_memory"))
        distances_value_error = Msg(__("msg.problem"))
        data_too_large_for_mahalanobis = Msg(__("msg.upto"))

    class Warning(OWWidget.Warning):
        ignoring_discrete = Msg("Ignoring categorical features")
        ignoring_nonbinary =  Msg(__("msg.ignoring_non-binary"))
        unsupported_sparse = Msg("Some metrics don't support sparse data\n"
                                 "and were disabled: {}")
        imputing_data = Msg(__("msg.miss_value"))
        no_features = Msg("Data has no features")

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.data = None

        gui.radioButtons(
            self.controlArea, self, "axis", [i18n.t("common.general.row"), i18n.t("common.general.column")],
            box=__("box_compare"), orientation=Qt.Horizontal, callback=self._invalidate
        )
        box = gui.hBox(self.controlArea, __("box_distance_metric"))
        self.metric_buttons = QButtonGroup()
        width = 0
        for i, metric in enumerate(MetricDefs.values()):
            if i % 6 == 0:
                vb = gui.vBox(box)
            b = QRadioButton(metric.name)
            b.setChecked(self.metric_id == metric.id)
            b.setToolTip(metric.tooltip)
            vb.layout().addWidget(b)
            width = max(width, b.sizeHint().width())
            self.metric_buttons.addButton(b, metric.id)
        for b in self.metric_buttons.buttons():
            b.setFixedWidth(width)

        self.metric_buttons.idClicked.connect(self._metric_changed)

        gui.auto_apply(self.buttonsArea, self, "autocommit")


    @Inputs.data
    @check_sql_input
    def set_data(self, data):
        self.cancel()
        self.data = data
        self.refresh_radios()
        self.commit.now()

    def _metric_changed(self, id_):
        self.metric_id = id_
        self._invalidate()

    def refresh_radios(self):
        sparse = self.data is not None and issparse(self.data.X)
        unsupported_sparse = []
        for metric in MetricDefs.values():
            button = self.metric_buttons.button(metric.id)
            no_sparse = sparse and not metric.metric.supports_sparse
            button.setEnabled(not no_sparse)
            if no_sparse:
                unsupported_sparse.append(metric.name)
        self.Warning.unsupported_sparse(", ".join(unsupported_sparse),
                                        shown=bool(unsupported_sparse))

    @gui.deferred
    def commit(self):
        self.compute_distances(self.data)

    def compute_distances(self, data):
        def _check_sparse():
            # pylint: disable=invalid-sequence-index
            if issparse(data.X) and not metric.supports_sparse:
                self.Error.dense_metric_sparse_data(metric_def.name)
                return False
            return True

        def _fix_discrete():
            nonlocal data
            if data.domain.has_discrete_attributes() \
                    and metric is not distance.Jaccard \
                    and (issparse(data.X) and getattr(metric, "fallback", None)
                         or not metric.supports_discrete
                         or self.axis == 1):
                if not data.domain.has_continuous_attributes():
                    self.Error.no_continuous_features()
                    return False
                self.Warning.ignoring_discrete()
                data = distance.remove_discrete_features(data, to_metas=True)
            return True

        def _fix_nonbinary():
            nonlocal data
            if metric is distance.Jaccard and not issparse(data.X):
                nbinary = sum(a.is_discrete and len(a.values) == 2
                              for a in data.domain.attributes)
                if not nbinary:
                    self.Error.no_binary_features()
                    return False
                elif nbinary < len(data.domain.attributes):
                    self.Warning.ignoring_nonbinary()
                    data = distance.remove_nonbinary_features(data,
                                                              to_metas=True)
            return True

        def _fix_missing():
            nonlocal data
            if not metric.supports_missing and bn.anynan(data.X):
                self.Warning.imputing_data()
                data = distance.impute(data)
            return True

        def _check_tractability():
            if metric is distance.Mahalanobis:
                if self.axis == 0:
                    # when computing distances by columns, we want < 1000 rows
                    if len(data) > 1000:
                        self.Error.data_too_large_for_mahalanobis("rows")
                        return False
                else:
                    if len(data.domain.attributes) > 1000:
                        self.Error.data_too_large_for_mahalanobis("columns")
                        return False
            return True

        def _check_no_features():
            if len(data.domain.attributes) == 0:
                self.Warning.no_features()
            return True

        metric_def = MetricDefs[self.metric_id]
        metric = metric_def.metric
        self.clear_messages()
        if data is not None:
            for check in (_check_sparse, _check_tractability,
                          _check_no_features,
                          _fix_discrete, _fix_missing, _fix_nonbinary):
                if not check():
                    data = None
                    break

        self.start(DistanceRunner.run, data, metric,
                   metric_def.normalize, self.axis)

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Orange.misc.DistMatrix):
        assert isinstance(result, Orange.misc.DistMatrix) or result is None
        self.Outputs.distances.send(result)

    def on_exception(self, ex):
        if isinstance(ex, ValueError):
            self.Error.distances_value_error(ex)
        elif isinstance(ex, MemoryError):
            self.Error.distances_memory_error()
        elif isinstance(ex, InterruptException):
            pass
        else:
            raise ex

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def _invalidate(self):
        self.commit.deferred()

    def send_report(self):
        # pylint: disable=invalid-sequence-index
        self.report_items((
            (__("row.distances_between"), [i18n.t("common.general.row"), __("row.columns")][self.axis]),
            (i18n.t("common.general.metric"), MetricDefs[self.metric_id].name)
        ))

    @classmethod
    def migrate_settings(cls, settings, version):
        if version is None or version < 2 and "normalized_dist" not in settings:
            # normalize_dist is set to False when restoring settings from
            # an older version to preserve old semantics.
            settings["normalized_dist"] = False
        if version is None or version < 3:
            # Mahalanobis was moved from idx = 2 to idx = 9
            metric_idx = settings["metric_idx"]
            if metric_idx == 2:
                settings["metric_idx"] = 9
            elif 2 < metric_idx <= 9:
                settings["metric_idx"] -= 1
        if version < 4:
            metric_idx = settings.pop("metric_idx")
            metric_id = [Euclidean, Manhattan, Cosine, Jaccard,
                         Spearman, SpearmanAbsolute, Pearson, PearsonAbsolute,
                         Hamming, Mahalanobis, Euclidean][metric_idx]
            if settings.pop("normalized_dist", False):
                metric_id = {Euclidean: EuclideanNormalized,
                             Manhattan: ManhattanNormalized}.get(metric_id,
                                                                 metric_id)
            settings["metric_id"] = metric_id


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWDistances).run(Orange.data.Table("iris"))



from bisect import bisect_left
from queue import Queue, Empty
from types import SimpleNamespace as namespace
from typing import Optional, Iterable, List, Callable, Iterator, Any, Type
from threading import Timer

from AnyQt.QtCore import (
    Qt, QSize, QSortFilterProxyModel, QObject, pyqtSignal as Signal)
from AnyQt.QtGui import (
    QStandardItemModel, QStandardItem, QShowEvent, QCloseEvent, QHideEvent)
from AnyQt.QtWidgets import (
    QTableView, QDialog, QVBoxLayout, QLineEdit, QPushButton)

from orangewidget.widget import OWBaseWidget
from Orange.widgets import gui
from Orange.widgets.gui import HorizontalGridDelegate, TableBarItem
from Orange.widgets.utils.concurrents import ConcurrentMixin, TaskState
from Orange.widgets.utils.progressbar import ProgressBarMixin


class Result(namespace):
    queue = None  # type: Queue[QueuedScore, ...]
    scores = None  # type: Optional[List[float, ...]]


class QueuedScore(namespace):
    position = None  # type: int
    score = None  # type: float
    state = None  # type: Iterable


class RunState(namespace):
    Invalid = 0  # Used only as default, changed to Initialized when instantiated
    Initialized = 1  # Has data; prepare_run must be called before starting computation
    Ready = 2  # Has data, iterator is initialized, but never used
    Running = 3  # Scoring thread is running
    Paused = 4  # Scoring thread is inactive, but can continue (without prepare_run)
    Hidden = 5  # Dialog is hidden and ranking paused; will resume on reopen
    Done = 6  # Scoring is done

    # The difference between Paused and Hidden is that for the former the
    # ranking is not automatically restarted when the dialog is reopened.

    state: int = Invalid
    iterator: Iterable = None
    completed: int = 0

    def can_run(self):
        return self.state in (self.Ready, self.Paused, self.Hidden)


class VizRankDialog(QDialog, ProgressBarMixin, ConcurrentMixin):
    """
    Base class for VizRank dialogs, providing a GUI with a table and a button,
    and the skeleton for managing the evaluation of visualizations.

    A new instance of this class is used for new data or any change made in
    the widget (e.g. color attribute in the scatter plot).

    The attribute run_state (and the related signal runStateChanged) can be
    used for tracking the work flow of the widget. run_state.state can be

    - RunState.Initialized (after __init__): the widget has the data, but
       the state generator is not constructed, total statecount is not known
       yet. The dialog may reenter this state after, for instance, changing
       the number of attributes per combination in radviz.
    - RunState.Ready (after prepare_run): the iterator is ready, state count
       is known: the ranking can commence.
    - RunState.Running (after start_computation): ranking is in progress.
       This state is entered by start_computation. If start_computation is
       called when the state is Initialized, start computation will call
       prepare_run.
    - RunState.Paused (after pause_computation): ranking is paused. This may
       continue (with start_computation) or be reset when parameters of VizRank
       are changed (by calling `set_run_state(RunState.Initialized)` and then
       start_computation, which will first call prepare_run to go to Ready).
    - RunState.Hidden (after closing the dialog): ranking is paused. This is
       similar to `RunState.Paused`, except that reopening dialog resumes it.
    - RunState.Done: ranking is done. The only allowed next state is
       Initialized (with the procedure described in the previous point).

    State should be changed through set_run_state, which also changes the button
    label and disables the button when Done.

    Derived classes must provide methods

    - `__init__(parent, *args, **kwargs)`: stores the data to run vizrank,
    - `state_generator()`: generates combinations (e.g. of attributes),
    - `compute_score(state)`: computes the score for the combination,
    - `row_for_state(state)`: returns a list of items inserted into the
       table for the given state.

    and, probably,

    - emit `selectionChanged` when the user changes the selection in the table,

    and, optionally,

    - `state_count`: return the number of combinations (used for progress bar)
    - `bar_length`: return the length of the bar corresponding to the score,
    - `auto_select`: selects the row corresponding to the given data.

    Derived classes are also responsible for connecting to
    rank_table.selectionModel().selectionChanged and emitting something useful
    via selectionChanged

    The constructor shouldn't do anything but store the necessary data (as is)
    because instances of this object are created on any new data. The actual
    work can only start in `prepare_run`. `prepare_run` usually won't be
    overriden, so the first computation in derived classes will usually happen
    in `compute_score` or, in classes derived from`VizRankAttributes`, in
    `score_attributes`.

    Args:
        parent (Orange.widget.OWWidget): widget to which the dialog belongs

    Attributes:
        captionTitle (str): the caption for the dialog. This can be a class
          attribute. `captionTitle` is used by the `ProgressBarMixin`.
        show_bars (True): if True (default), table with scores contains bars
          of length -score. For a different length, override `bar_lenght`.
          To hide bars (e.g. because scores can't be normalized) set this to
          `False`.

    Signal:
        selectionChanged(object): emitted when selection in the table is
            changed. The data type depends on the derived class (e.g. a list
            of attributes)
        runStateChanged(int, dict): emitted when the run state changes
            (e.g. start, pause...). Derived classes can fill the dictionary
            with additional data, e.g. the state of the user interface
    """
    captionTitle = "Score Plots"
    show_bars = True

    selectionChanged = Signal(object)
    runStateChanged = Signal(int, dict)

    button_labels = {RunState.Initialized: "Start",
                     RunState.Ready: "Start",
                     RunState.Running: "Pause",
                     RunState.Paused: "Continue",
                     RunState.Hidden: "Continue",
                     RunState.Done: "Finished",
                     RunState.Invalid: "Start"}

    def __init__(self, parent):
        QDialog.__init__(self, parent, windowTitle=self.captionTitle)
        ConcurrentMixin.__init__(self)
        ProgressBarMixin.__init__(self)
        self.setLayout(QVBoxLayout())

        self.scores = []
        self.add_to_model = Queue()
        self.run_state = RunState()
        self.total_states = 1

        self.filter = QLineEdit()
        self.filter.setPlaceholderText("Filter ...")
        self.layout().addWidget(self.filter)

        self.rank_model = QStandardItemModel(self)
        self.model_proxy = QSortFilterProxyModel(
            self, filterCaseSensitivity=Qt.CaseInsensitive)
        self.model_proxy.setSourceModel(self.rank_model)
        self.filter.textChanged.connect(self.model_proxy.setFilterFixedString)

        self.rank_table = view = QTableView(
            selectionBehavior=QTableView.SelectRows,
            selectionMode=QTableView.SingleSelection,
            showGrid=False,
            editTriggers=gui.TableView.NoEditTriggers)
        view.setItemDelegate(TableBarItem() if self.show_bars
                             else HorizontalGridDelegate())
        view.setModel(self.model_proxy)
        view.horizontalHeader().setStretchLastSection(True)
        view.horizontalHeader().hide()
        self.layout().addWidget(view)

        self.button = gui.button(self, self, "Start", default=True)

        @self.button.pressed.connect
        def on_button_pressed():
            if self.run_state.state == RunState.Running:
                self.pause_computation()
            else:
                self.start_computation()

        self.set_run_state(RunState.Initialized)

    def prepare_run(self) -> None:
        """
        Called by start_computation before running for the first time or with
        new parameters within the vizrank gui, e.g. a different number of
        attributes in combinations.

        Derived classes may override this method to add other preparation steps,
        but shouldn't need to call it.
        """
        self.progressBarInit()
        self.scores = []
        self._update_model()  # empty queue
        self.rank_model.clear()
        self.run_state.iterator = self.state_generator()
        self.total_states = self.state_count() or 1
        self.set_run_state(RunState.Ready)

    def start_computation(self) -> None:
        if self.run_state.state == RunState.Initialized:
            self.prepare_run()
        if not self.run_state.can_run():
            return
        self.set_run_state(RunState.Running)
        self.start(
            self.run_vizrank,
            self.compute_score, self.scores,
            self.run_state.iterator, self.run_state.completed)

    def pause_computation(self, new_state=RunState.Paused) -> None:
        if not self.run_state.state == RunState.Running:
            return
        self.set_run_state(new_state)
        self.cancel()
        self._update_model()

    def dialog_reopened(self):
        if self.run_state.state != RunState.Paused:
            self.start_computation()

    @staticmethod
    def run_vizrank(compute_score: Callable, scores: List,
                    state_iterator: Iterator, completed: int, task: TaskState):
        res = Result(queue=Queue(), scores=None, completed_states=completed)
        scores = scores.copy()
        can_set_partial_result = True

        def do_work(st: Any):
            score = compute_score(st)
            if score is not None:
                pos = bisect_left(scores, score)
                res.queue.put_nowait(QueuedScore(position=pos, score=score,
                                                 state=st))
                scores.insert(pos, score)
            res.scores = scores.copy()
            res.completed_states += 1

        def reset_flag():
            nonlocal can_set_partial_result
            can_set_partial_result = True

        for state in state_iterator:
            do_work(state)
            # Prevent simple scores from invoking 'task.set_partial_result')
            # too frequently and making the widget unresponsive
            if can_set_partial_result:
                task.set_partial_result(res)
                can_set_partial_result = False
                Timer(0.01, reset_flag).start()
            if task.is_interruption_requested():
                return res
        task.set_partial_result(res)
        return res

    def on_partial_result(self, result: Result) -> None:
        try:
            while True:
                queued = result.queue.get_nowait()
                self.add_to_model.put_nowait(queued)
        except Empty:
            pass
        self.scores = result.scores
        self._update_model()
        self.run_state.completed = result.completed_states
        self.progressBarSet(self._progress)

    @property
    def _progress(self) -> int:
        return int(round(self.run_state.completed * 100 / self.total_states))

    def on_done(self, result: Result) -> None:
        self.progressBarFinished()
        self.set_run_state(RunState.Done)
        self._update_model()
        if not self.rank_table.selectedIndexes() \
                and self.rank_table.model().rowCount():
            self.rank_table.selectRow(0)

    def set_run_state(self, state: Any) -> None:
        if state != self.run_state.state:
            self.run_state.state = state
            if state <= RunState.Ready:
                self.run_state.completed = 0
            self.emit_run_state_changed()
        if state == RunState.Paused:
            self.setWindowTitle(
                f"{self.captionTitle} (paused at {self._progress}%)")
        self.set_button_state()

    def emit_run_state_changed(self):
        self.runStateChanged.emit(self.run_state.state, {})

    def set_button_state(self,
                         label: Optional[str] = None,
                         enabled: Optional[bool]=None) -> None:
        state = self.run_state.state
        self.button.setText(
            label if label is not None
            else self.button_labels[state])
        self.button.setEnabled(
            enabled if enabled is not None
            else state not in [RunState.Done, RunState.Invalid])

    def _update_model(self) -> None:
        try:
            while True:
                queued = self.add_to_model.get_nowait()
                row_items = self.row_for_state(queued.score, queued.state)
                bar_length = self.bar_length(queued.score)
                if bar_length is not None:
                    row_items[0].setData(bar_length,
                                         gui.TableBarItem.BarRole)
                self.rank_model.insertRow(queued.position, row_items)
        except Empty:
            pass

    def showEvent(self, event: QShowEvent) -> None:
        # pylint: disable=protected-access
        self.parent()._restore_vizrank_geometry()
        super().showEvent(event)

    def hideEvent(self, event: QHideEvent) -> None:
        # pylint: disable=protected-access
        self.pause_computation(RunState.Hidden)
        self.parent()._save_vizrank_geometry()
        super().hideEvent(event)

    def state_generator(self) -> Iterable:
        """
        Generate all possible states (e.g. attribute combinations) for the
        given data. The content of the generated states is specific to the
        visualization.
        """
        raise NotImplementedError

    def compute_score(self, state: Any) -> Optional[float]:
        """
        Abstract method for computing the score for the given state. Smaller
        scores are better.

        Args:
            state: the state, e.g. the combination of attributes as generated
                by :obj:`state_count`.
        """
        raise NotImplementedError

    def row_for_state(self, score: float, state: Any) -> List[QStandardItem]:
        """
        Return a list of items that are inserted into the table.

        Args:
            score: score, computed by :obj:`compute_score`
            state: the state, e.g. combination of attributes
            """
        raise NotImplementedError

    def auto_select(self, arg: Any) -> None:
        """
        Select the row corresponding to the give data.
        """

    def state_count(self) -> int:
        """
        Return the total number of states, needed for the progress bar.
        """
        return 1

    def bar_length(self, score: float) -> float:
        """
        Return the bar length (between 0 and 1) corresponding to the score.
        Return `None` if the score cannot be normalized.
        """
        return max(0., -score)


# According to PEP-8, names should follow usage, not implementation
# pylint: disable=invalid-name
def VizRankMixin(vizrank_class: Type[VizRankDialog]) -> Type[type]:
    """
    A mixin that serves as an interface between the vizrank dialog and the widget.

    Widget should avoid directly access the vizrank dialog.

    This mixin takes care of constructing the vizrank dialog, raising it,
    and for closing and shutting the vizrank down when necessary. Data for
    vizrank is passed to the vizrank through the mixin, and the mixin forwards
    the signals from vizrank dialog (e.g. when the user selects rows in vizrank)
    to the widget.

    The mixin is parametrized: it must be given the VizRank class to open,
    as in

    ```
    class OWMosaicDisplay(OWWidget, VizRankMixin(MosaicVizRank)):
    ```

    There should therefore be no need to subclass this class.

    Method `vizrank_button` returns a button to be placed into the widget.

    Signals:
     - `vizrankSelectionChanged` is connected to VizRank's selectionChanged.
       E.g. MosaicVizRank.selectionChanged is forwarded to selectionChanged,
       and contains the data sent by the former.
     - `virankRunStateChanged` is connected to VizRank's runStateChanged.
       This can be used to retrieve the settings from the dialog at appropriate
       times.
    - If the widget emits a signal `vizrankAutoSelect` when
      the user manually changes the variables shown in the plot (e.g. x or y
      attribute in the scatter plot), the vizrank will also select this
      combination in the list, if found.
"""

    # __VizRankMixin must be derived from OWBaseWidget so that it is
    # properly included in MRO and calls to super(),
    # e.g. super().onDeleteWidget() are properly called.
    #
    # Yet PyQt cannot handle diamond-shaped inheritance: signals exist,
    # but can't be connected. While PyQt states that "It is not possible to
    # define a new Python class that sub-classes from more than one Qt class
    # (https://www.riverbankcomputing.com/static/Docs/PyQt5/gotchas.html#multiple-inheritance),
    # this is not the case here, because we subclass from QDialog in the
    # tip of the diamond, hence it is a single inheritance with multiple
    # paths. Hence we define signals in a separate object, instantiate it,
    # and forward the signals through it.
    class VizrankSignalDelegator(QObject):
        vizrankSelectionChanged = Signal(object)
        vizrankRunStateChanged = Signal(int, dict)
        vizrankAutoSelect = Signal(object)

    class __VizRankMixin(OWBaseWidget, openclass=True):  # pylint: disable=invalid_name
        __button: Optional[QPushButton] = None
        __vizrank: Optional[VizRankDialog] = None
        __vizrank_geometry: Optional[bytes] = None

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__signal_delegator = VizrankSignalDelegator()

        @property
        def vizrankSelectionChanged(self):
            return self.__signal_delegator.vizrankSelectionChanged

        @property
        def vizrankAutoSelect(self):
            return self.__signal_delegator.vizrankAutoSelect

        @property
        def vizrankRunStateChanged(self):
            return self.__signal_delegator.vizrankRunStateChanged

        @property
        def vizrank_dialog(self) -> Optional[VizRankDialog]:
            """
            The vizrank dialog. This should be used only in tests.
            """
            return self.__vizrank

        def vizrank_button(self, button_label: Optional[str] = None) -> QPushButton:
            """
            A button that opens/starts the vizrank.

            The label is optional because this function is used for
            constructing the button as well as for retrieving it later.
            """
            if self.__button is None:
                self.__button = QPushButton()
                self.__button.pressed.connect(self.start_vizrank)
                self.__button.pressed.connect(self.raise_vizrank)
                # It's implausible that vizrank_button will be called after
                # init_vizrank, we could just disable the button. But let's
                # play it safe.
                self.__button.setDisabled(self.__vizrank is None)
            if button_label is not None:
                self.__button.setText(button_label)
            return self.__button

        def init_vizrank(self, *args, **kwargs) -> None:
            """
            Construct the vizrank dialog

            Any data is given to the constructor. This also enables the button
            if it exists.
            """
            self.shutdown_vizrank()
            self.__vizrank = vizrank_class(self, *args, **kwargs)
            self.__vizrank.selectionChanged.connect(self.vizrankSelectionChanged)
            self.__vizrank.runStateChanged.connect(self.vizrankRunStateChanged)
            self.vizrankAutoSelect.connect(self.__vizrank.auto_select)
            # There may be Vizrank without a button ... perhaps.
            if self.__button is not None:
                self.__button.setEnabled(True)
                self.__button.setToolTip("")

        def disable_vizrank(self, reason: str = "") -> None:
            """
            Shut down the vizrank thread, closes it, disables the button.

            The method should be called when the widget has no data or cannot
            run vizrank on it. The optional `reason` is set as tool tip.
            """
            self.shutdown_vizrank()
            if self.__button is not None:
                self.__button.setEnabled(False)
                self.__button.setToolTip(reason)

        def start_vizrank(self) -> None:
            """
            Start the ranking.

            There should be no reason to call this directly,
            unless the widget has no vizrank button.
            """
            self.__vizrank.dialog_reopened()

        def raise_vizrank(self) -> None:
            """
            Start the ranking.

            There should be no reason to call this directly,
            unless the widget has no vizrank button.
            """
            if self.__vizrank is None:
                return
            self.__vizrank.show()
            self.__vizrank.activateWindow()
            self.__vizrank.raise_()

        def shutdown_vizrank(self) -> None:
            """
            Start the ranking.

            There should be no reason to call this directly:
            the method is called
            - from init_vizrank (the widget received new data),
            - from disable_vizrank (the widget lost data, or data is unsuitable),
            - when the widget is deleted.
            """
            if self.__vizrank is None:
                return
            self.__vizrank.shutdown()
            self.__vizrank.close()
            self.__vizrank.deleteLater()
            self.__vizrank = None

        # The following methods ensure that the vizrank dialog is
        # closed/hidden/destroyed together with its parent widget.
        def closeEvent(self, event: QCloseEvent) -> None:
            if self.__vizrank:
                self.__vizrank.close()
            super().closeEvent(event)

        def hideEvent(self, event: QHideEvent) -> None:
            if self.__vizrank:
                self.__vizrank.hide()
            super().hideEvent(event)

        def onDeleteWidget(self) -> None:
            self.shutdown_vizrank()
            super().onDeleteWidget()

        def _save_vizrank_geometry(self) -> None:
            assert self.__vizrank
            self.__vizrank_geometry = self.__vizrank.saveGeometry()

        def _restore_vizrank_geometry(self) -> None:
            assert self.__vizrank
            if self.__vizrank_geometry is not None:
                self.__vizrank.restoreGeometry(self.__vizrank_geometry)

    # Give the returned class a proper name, like MosaicVizRankMixin
    return type(f"{vizrank_class.__name__}Mixin", (__VizRankMixin, ), {})


# This is an abstract class, pylint: disable=abstract-method
class VizRankDialogAttrs(VizRankDialog):
    """
    Base class for VizRank classes that work over combinations of attributes.

    Constructor accepts
     - data (Table),
     - attributes (list[Variable]; if omitted, data.domain.variables is used),
     - attr_color (Optional[Variable]): the "color" attribute, if applicable.

    The class assumes that `state` is a sequence of indices into a list of
    attributes. On this basis, it provides

    - `row_for_state`, that constructs a list containing a single QStandardItem
      with names of attributes and with `_AttrRole` data set to a list of
      attributes for the given state.
    - `on_selection_changed` that emits a `selectionChanged` signal with the
      above list.

    Derived classes must still provide

    - `state_generator()`: generates combinations of attribute indices
    - `compute_score(state)`: computes the score for the combination

    Derived classes will usually provide
    - `score_attribute` that will returned a list of attributes, such as found
      in self.attrs, but sorted by importance according to some heuristic.

    Attributes:
        - data (Table): data used in ranking
        - attrs (list[Variable]): applicable variables
        - attr_color (Variable or None): the target attribute

    Class attributes:
        - sort_names_in_row (bool): if set to True (default is False),
          variables in the view will be sorted alphabetically.
    """
    _AttrRole = next(gui.OrangeUserRole)
    sort_names_in_row = False

    # Ideally, the only argument would be `data`, with attributes used for
    # ranking and class_var would be the "color". This would however require
    # that widgets prepare such data when initializing vizrank even though
    # vizrank wouldn't necessarily be called at all. We will be able to afford
    # that after we migrate Table to pandas.
    def __init__(self, parent,
                 data: "Orange.data.Table",
                 attributes: Optional[List["Orange.data.Variable"]] = None,
                 attr_color: Optional["Orange.data.Variable"] = None):
        super().__init__(parent)
        self.data = data or None
        self.attrs = attributes or (
            self.data and self.data.domain.variables)
        self.attr_color = attr_color
        self._attr_order = None

        self.rank_table.selectionModel().selectionChanged.connect(
            self.on_selection_changed)

    @property
    def attr_order(self) -> List["Orange.data.Variable"]:
        """
        Attributes, sorted according to some heuristic.

        The property is computed by score_attributes when neceessary, and
        cached.
        """
        if self._attr_order is None:
            self._attr_order = self.score_attributes()
        return self._attr_order

    def score_attributes(self) -> None:
        """
        Return a list of attributes ordered according by some heuristic.

        Default implementation returns the original list `self.attrs`.
        """
        return self.attrs

    def on_selection_changed(self, selected, deselected) -> None:
        """
        Emit the currently selected combination of variables.
        """
        selection = selected.indexes()
        if not selection:
            return
        attrs = selected.indexes()[0].data(self._AttrRole)
        self.selectionChanged.emit(attrs)

    def row_for_state(self, score: float, state: List[int]
                      ) -> List[QStandardItem]:
        """
        Return the QStandardItem for the given combination of attributes.
        """
        attrs = [self.attr_order[s] for s in state]
        if self.sort_names_in_row:
            attrs.sort(key=lambda attr: attr.name.lower())
        attr_names = (a.name for a in attrs)
        item = QStandardItem(', '.join(attr_names))
        item.setData(attrs, self._AttrRole)
        return [item]

    # Renamed from the more general 'arg', pylint: disable=arguments-renamed
    def auto_select(self, attrs: List["Orange.data.Variable"]) -> None:
        """
        Find the given combination of variables and select it (if it exists)
        """
        model = self.rank_model
        self.rank_table.selectionModel().clear()
        for row in range(model.rowCount()):
            index = model.index(row, 0)
            row_attrs = model.data(index, self._AttrRole)
            if all(x is y for x, y in zip(row_attrs, attrs)):
                self.rank_table.selectRow(row)
                self.rank_table.scrollTo(index)
                return


# This is an abstract class, pylint: disable=abstract-method
class VizRankDialogAttrPair(VizRankDialogAttrs):
    """
    Base class for VizRanks with combinations of two variables.

    Provides state_generator and state_count; derived classes must provide
    compute_score and, possibly, score_attributes.
    """
    def __init__(self, parent, data, attributes=None, attr_color=None):
        super().__init__(parent, data, attributes, attr_color)
        self.resize(320, 512)

    def sizeHint(self) -> QSize:
        return QSize(320, 512)

    def state_count(self) -> int:
        n_attrs = len(self.attrs)
        return n_attrs * (n_attrs - 1) // 2

    def state_generator(self) -> Iterable:
        return ((j, i) for i in range(len(self.attr_order)) for j in range(i))


# This is an abstract class, pylint: disable=abstract-method
class VizRankDialogNAttrs(VizRankDialogAttrs):
    """
    Base class for VizRanks with a spin for selecting the number of attributes.

    Constructor requires data, attributes, attr_color and, also, the initial
    number of attributes in the spin box.

    - Ranking is stopped if the user interacts with the spin.
    - The button label is changed to "Restart with {...} variables" if the
      number selected in the spin doesn't match the number of varialbes in the
      paused ranking, and reset back to "Continue" when it matches.
    - start_computation is overriden to lower the state to Initialized before
      calling super, if the number of selected in the spin doesn't match the
      previous run.
    - The dictionary passed by the signal runStateChanged contains n_attrs
      with the number of attributes used in the current/last ranking.
    - When closing the dialog, the spin is reset to the number of attributes
      used in the last run.
    """
    attrsSelected = Signal([])

    def __init__(self, parent,
                 data: "Orange.data.Table",
                 attributes: List["Orange.data.Variable"],
                 color: "Orange.data.Variable",
                 n_attrs: int,
                 *, spin_label: str = "Number of variables: "):
        # Add the spin box for a number of attributes to take into account.
        self.n_attrs = n_attrs
        super().__init__(parent, data, attributes, color)
        self._attr_order = None

        box = gui.hBox(self)
        self.n_attrs_spin = gui.spin(
            box, self, None, 3, 8, label=spin_label,
            controlWidth=50, alignment=Qt.AlignRight,
            callback=self.on_n_attrs_changed)

        n_cont = self.max_attrs()
        self.n_attrs_spin.setValue(min(self.n_attrs, n_cont))
        self.n_attrs_spin.setMaximum(n_cont)
        self.n_attrs_spin.parent().setDisabled(not n_cont)

    def max_attrs(self) -> int:
        return sum(v is not self.attr_color for v in self.attrs)

    def start_computation(self) -> None:
        if self.n_attrs != self.n_attrs_spin.value():
            self.n_attrs = self.n_attrs_spin.value()
            self.set_run_state(RunState.Initialized)
        self.n_attrs_spin.lineEdit().deselect()
        self.rank_table.setFocus(Qt.FocusReason.OtherFocusReason)
        super().start_computation()

    def on_n_attrs_changed(self) -> None:
        if self.run_state.state == RunState.Running:
            self.pause_computation()

        new_attrs = self.n_attrs_spin.value()
        if new_attrs == self.n_attrs:
            self.set_button_state()
        else:
            self.set_button_state(label=f"Restart with {new_attrs} variables",
                                  enabled=True)

    def emit_run_state_changed(self) -> None:
        self.runStateChanged.emit(self.run_state.state,
                                  {"n_attrs":self.n_attrs})

    def closeEvent(self, event) -> None:
        self.n_attrs_spin.setValue(self.n_attrs)
        super().closeEvent(event)



import numpy as np
from AnyQt.QtCore import Qt
import scipy.sparse as sp

from Orange.data import Table, Domain, ContinuousVariable, StringVariable, \
    DiscreteVariable
from Orange.regression import PLSRegressionLearner
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.signals import Output
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Msg
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.model.model.owpls." + key)

class OWPLS(OWBaseLearner):
    name = 'PLS'
    description = "Partial Least Squares Regression widget for multivariate data analysis"
    icon = "icons/PLS.svg"
    priority = 85
    keywords = ["partial least squares"]

    LEARNER = PLSRegressionLearner

    class Outputs(OWBaseLearner.Outputs):
        coefsdata = Output("Coefficients and Loadings", Table, explicit=True)
        data = Output("Data with Scores", Table)
        components = Output("Components", Table)

    class Warning(OWBaseLearner.Warning):
        sparse_data = Msg(
            'Sparse input data: default preprocessing is to scale it.')

    n_components = Setting(2)
    max_iter = Setting(500)
    scale = Setting(True)

    def add_main_layout(self):
        optimization_box = gui.vBox(
            self.controlArea,__("optimization_parameters"))
        gui.spin(
            optimization_box, self, "n_components", 1, 50, 1,
            label=__("components"),
            alignment=Qt.AlignRight, controlWidth=100,
            callback=self.settings_changed)
        gui.spin(
            optimization_box, self, "max_iter", 5, 1000000, 50,
            label=__("iteration_limit"),
            alignment=Qt.AlignRight, controlWidth=100,
            callback=self.settings_changed,
            checkCallback=self.settings_changed)
        gui.checkBox(optimization_box, self, "scale",
                     __("scale_features_and_target"),
                     callback=self.settings_changed)

    def update_model(self):
        super().update_model()
        coef_table = None
        data = None
        components = None
        if self.model is not None:
            coef_table = self._create_output_coeffs_loadings()
            data = self._create_output_data()
            components = self.model.components()
        self.Outputs.coefsdata.send(coef_table)
        self.Outputs.data.send(data)
        self.Outputs.components.send(components)

    def _create_output_coeffs_loadings(self) -> Table:
        intercept = self.model.intercept.T[None, :]
        coefficients = self.model.coefficients.T
        _, y_loadings = self.model.loadings
        x_rotations, _ = self.model.rotations

        n_features, n_targets = coefficients.shape
        n_components = x_rotations.shape[1]

        names = [f"coef ({v.name})" for v in self.model.domain.class_vars]
        names += [f"coef/X_sd ({v.name})" for v in self.model.domain.class_vars]
        names += [f"w*c {i + 1}" for i in range(n_components)]
        domain = Domain(
            [ContinuousVariable(n) for n in names],
            metas=[StringVariable("Variable name"),
                   DiscreteVariable("Variable role", ("Feature", "Target"))]
        )

        data = self.model.data_to_model_domain(self.data)
        x_std = np.std(data.X, axis=0)
        coeffs_x_std = coefficients.T / x_std
        X_features = np.hstack((coefficients,
                                coeffs_x_std.T,
                                x_rotations))
        X_targets = np.hstack((np.full((n_targets, n_targets), np.nan),
                               np.full((n_targets, n_targets), np.nan),
                               y_loadings))

        coeffs = coeffs_x_std * np.mean(data.X, axis=0)
        X_intercepts = np.hstack((intercept,
                                  intercept - coeffs.sum(),
                                  np.full((1, n_components), np.nan)))
        X = np.vstack((X_features, X_targets, X_intercepts))

        variables = self.model.domain.variables
        M = np.array([[v.name for v in variables] + ["intercept"],
                      [0] * n_features + [1] * n_targets + [np.nan]],
                     dtype=object).T

        table = Table.from_numpy(domain, X=X, metas=M)
        table.name = "Coefficients and Loadings"
        return table

    def _create_output_data(self) -> Table:
        projection = self.model.project(self.data)
        normal_probs = self.model.residuals_normal_probability(self.data)
        dmodx = self.model.dmodx(self.data)
        data_domain = self.data.domain
        proj_domain = projection.domain
        nprobs_domain = normal_probs.domain
        dmodx_domain = dmodx.domain
        metas = data_domain.metas + proj_domain.attributes + proj_domain.metas + \
            nprobs_domain.attributes + dmodx_domain.attributes
        domain = Domain(data_domain.attributes, data_domain.class_vars, metas)
        data: Table = self.data.transform(domain)
        with data.unlocked(data.metas):
            data.metas[:, -2 * len(self.data.domain.class_vars) - 1: -1] = \
                normal_probs.X
            data.metas[:, -1] = dmodx.X[:, 0]
        return data

    @OWBaseLearner.Inputs.data
    def set_data(self, data):
        # reimplemented completely because the base learner does not
        # allow multiclass

        self.Warning.sparse_data.clear()

        self.Error.data_error.clear()
        self.data = data

        if data is not None and data.domain.class_var is None and not data.domain.class_vars:
            self.Error.data_error(
                "Data has no target variable.\n"
                "Select one with the Select Columns widget.")
            self.data = None

        # invalidate the model so that handleNewSignals will update it
        self.model = None

        if self.data and sp.issparse(self.data.X):
            self.Warning.sparse_data()

    def create_learner(self):
        common_args = {'preprocessors': self.preprocessors}
        return PLSRegressionLearner(n_components=self.n_components,
                                    scale=self.scale,
                                    max_iter=self.max_iter,
                                    **common_args)


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWPLS).run(Table("housing"))



from itertools import islice, permutations, chain, combinations
from math import factorial, comb
import warnings

import numpy as np
import pyqtgraph as pg
from AnyQt.QtCore import Qt, QRectF, QPoint, pyqtSignal as Signal
from AnyQt.QtGui import QStandardItem, QColor

import pyqtgraph as pg
from pyqtgraph.graphicsItems.ScatterPlotItem import ScatterPlotItem
from scipy.spatial import distance
from sklearn.model_selection import cross_val_score
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from Orange.data import Table, Domain
from AnyQt.QtGui import QColor, QPalette
from AnyQt.QtCore import Qt, QRectF, QPoint

import pyqtgraph as pg
from pyqtgraph.graphicsItems.ScatterPlotItem import ScatterPlotItem

from Orange.data import Table, Domain, IsDefined
from Orange.preprocess.score import ReliefF, RReliefF
from Orange.projection import RadViz
from Orange.widgets import widget, gui
from Orange.widgets.settings import Setting, ContextSetting, SettingProvider
from Orange.widgets.utils.plot.owplotgui import VariableSelectionModel, \
    variables_selection
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.visualize.utils import vizrank
from Orange.widgets.visualize.utils.vizrank import VizRankDialogNAttrs, \
    VizRankMixin
from Orange.widgets.visualize.utils.component import OWGraphWithAnchors
from Orange.widgets.visualize.utils.plotutils import TextItem
from Orange.widgets.visualize.utils.widget import OWAnchorProjectionWidget
from Orange.i18n_config import *

MAX_DISPLAYED_VARS = 20
MAX_LABEL_LEN = 16

def __(key):
    return i18n.t("widget.visualize.visualize.owradviz." + key)
class RadvizVizRank(VizRankDialogNAttrs):
    minK = 10

    def __init__(self, parent, data, attributes, color, n_attrs):
        super().__init__(parent, data, attributes, color, n_attrs,
                         spin_label="Maximum number of variables: ")

    def score_attributes(self):
        attrs = [v for v in self.attrs if v is not self.attr_color]
        data = self.data.transform(Domain(attrs, self.attr_color))
        relief = ReliefF if self.attr_color.is_discrete else RReliefF
        weights = relief(n_iterations=100, k_nearest=self.minK)(data)
        attrs = sorted(zip(weights, attrs), key=lambda x: (-x[0], x[1].name))
        return [a for _, a in attrs]

    def state_count(self):
        n_all_attrs = self.max_attrs()
        if not n_all_attrs:
            return 0
        return sum(comb(n_all_attrs, n_attrs) * factorial(n_attrs - 1) // 2
                   for n_attrs in range(3, self.n_attrs + 1))

    def state_generator(self):
        return (
            (c[0], *p)
            for k in range(3, self.n_attrs + 1)
            for c in combinations(list(range(len(self.attr_order))), k)
            for p in islice(permutations(c[1:]), factorial(len(c) - 1) // 2))

    def compute_score(self, state):
        attrs = [self.attr_order[i] for i in state]
        domain = Domain(attributes=attrs, class_vars=[self.attr_color])
        data = self.data.transform(domain)
        valid_data = IsDefined()(data)
        projector = RadViz()
        projection = projector(valid_data)
        radviz_xy = projection(valid_data).X
        y = projector.preprocess(valid_data).Y

        neigh = (KNeighborsClassifier if self.attr_color.is_discrete else
                 KNeighborsRegressor)(n_neighbors=3)
        neigh.fit(radviz_xy, y)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            scores = cross_val_score(neigh, radviz_xy, y, cv=3)
        return -scores.mean() * len(valid_data) / len(data)


class OWRadvizGraph(OWGraphWithAnchors):
    def __init__(self, scatter_widget, parent):
        super().__init__(scatter_widget, parent)
        self.anchors_scatter_item = None
        self.padding = 0.025

    def clear(self):
        super().clear()
        self.anchors_scatter_item = None

    def set_view_box_range(self):
        self.view_box.setRange(QRectF(-1, -1, 2, 2), padding=self.padding)

    def closest_draggable_item(self, pos):
        points, _ = self.master.get_anchors()
        if points is None:
            return None
        np_pos = np.array([[pos.x(), pos.y()]])
        distances = distance.cdist(np_pos, points[:, :2])[0]
        if len(distances) and np.min(distances) < self.DISTANCE_DIFF:
            return np.argmin(distances)
        return None

    def update_anchors(self):
        points, labels = self.master.get_anchors()
        if points is None:
            return
        if self.anchor_items is not None:
            for anchor in self.anchor_items:
                self.plot_widget.removeItem(anchor)

        self.anchor_items = []
        label_len = 1
        foreground = self.plot_widget.palette().color(QPalette.Text)
        for point, label in zip(points, labels):
            anchor = TextItem(color=foreground)
            anchor.textItem.setToolTip(f"<b>{label}</b>")

            if len(label) > MAX_LABEL_LEN:
                i = label.rfind(" ", 0, MAX_LABEL_LEN)
                if i != -1:
                    first_row = label[:i] + "\n"
                    second_row = label[i + 1:]
                    if len(second_row) > MAX_LABEL_LEN:
                        j = second_row.rfind(" ", 0, MAX_LABEL_LEN)
                        if j != -1:
                            second_row = second_row[:j + 1] + "..."
                        else:
                            second_row = second_row[:MAX_LABEL_LEN - 3] + "..."
                    label = first_row + second_row
                else:
                    label = label[:MAX_LABEL_LEN - 3] + "..."

            anchor.setText(label)
            anchor.setFont(self.parameter_setter.anchor_font)
            label_len = min(MAX_LABEL_LEN, len(label))

            x, y = point
            angle = np.rad2deg(np.arctan2(y, x))
            anchor.setPos(x * 1.025, y * 1.025)

            if abs(angle) < 90:
                anchor.setAngle(angle)
                anchor.setAnchor((0, 0.5))
            else:
                anchor.setAngle(angle + 180)
                anchor.setAnchor((1, 0.5))

                anchor.textItem.setTextWidth(anchor.textItem.boundingRect().width())
                option = anchor.textItem.document().defaultTextOption()
                option.setAlignment(Qt.AlignRight)
                anchor.textItem.document().setDefaultTextOption(option)

            self.plot_widget.addItem(anchor)
            self.anchor_items.append(anchor)

        self.padding = label_len * 0.0175
        self._update_anchors_scatter_item(points)

    def _update_anchors_scatter_item(self, points):
        if self.anchors_scatter_item is not None:
            self.plot_widget.removeItem(self.anchors_scatter_item)
            self.anchors_scatter_item = None
        self.anchors_scatter_item = ScatterPlotItem(x=points[:, 0],
                                                    y=points[:, 1])
        self.plot_widget.addItem(self.anchors_scatter_item)

    def _add_indicator_item(self, anchor_idx):
        if anchor_idx is None:
            return
        x, y = self.anchor_items[anchor_idx].get_xy()
        col = self.view_box.mouse_state
        dx = (self.view_box.childGroup.mapToDevice(QPoint(1, 0)) -
              self.view_box.childGroup.mapToDevice(QPoint(-1, 0))).x()
        self.indicator_item = MoveIndicator(np.arctan2(y, x), col, 6000 / dx)
        self.plot_widget.addItem(self.indicator_item)


class OWRadviz(OWAnchorProjectionWidget, VizRankMixin(RadvizVizRank)):
    name = __("name")
    description = __("desc")
    icon = "icons/Radviz.svg"
    priority = 241
    keywords = "radviz, viz"

    settings_version = 3

    selected_vars = ContextSetting([])
    GRAPH_CLASS = OWRadvizGraph
    graph = SettingProvider(OWRadvizGraph)
    n_attrs_vizrank = Setting(3)

    class Warning(OWAnchorProjectionWidget.Warning):
        # invalid_embedding = widget.Msg(__("msg.no_projection"))
        removed_vars = widget.Msg(__("msg.show_error"))
        max_vars_selected = widget.Msg(__("msg.number_error"))

    def __init__(self):
        VizRankMixin.__init__(self)  # pylint: disable=non-parent-init-called
        OWAnchorProjectionWidget.__init__(self)

    def _add_controls(self):
        box = gui.vBox(self.controlArea, box=i18n.t("common.general.features"))
        self.model_selected = VariableSelectionModel(self.selected_vars,
                                                     max_vars=20)
        variables_selection(box, self, self.model_selected)
        self.model_selected.selection_changed.connect(
            self.__model_selected_changed)
        box.layout().addWidget(self.vizrank_button(__("btn.suggest_feature")))
        # box.layout().addWidget(self.vizrank_button("Suggest features"))
        self.vizrankSelectionChanged.connect(self.vizrank_set_attrs)
        self.vizrankRunStateChanged.connect(self.store_vizrank_n_attrs)
        super()._add_controls()

    def _add_buttons(self):
        self.gui.box_zoom_select(self.buttonsArea)
        gui.auto_send(self.buttonsArea, self, "auto_commit")

    @property
    def primitive_variables(self):
        if self.data is None or self.data.domain is None:
            return []
        dom = self.data.domain
        return [v for v in chain(dom.variables, dom.metas)
                if v.is_continuous or v.is_discrete and len(v.values) == 2]

    @property
    def effective_variables(self):
        return self.selected_vars

    @property
    def effective_data(self):
        return self.data.transform(Domain(self.effective_variables))

    def store_vizrank_n_attrs(self, state, data):
        if state == vizrank.RunState.Running:
            self.n_attrs_vizrank = data["n_attrs"]

    def vizrank_set_attrs(self, attrs):
        if not attrs:
            return
        self.selected_vars[:] = attrs
        # Ugly, but the alternative is to have yet another signal to which
        # the view will have to connect
        self.model_selected.selection_changed.emit()

    def __model_selected_changed(self):
        if self.model_selected.is_full():
            self.Warning.max_vars_selected()
        else:
            self.Warning.max_vars_selected.clear()
        self.init_projection()
        self.setup_plot()
        self.commit.deferred()

    def colors_changed(self):
        super().colors_changed()
        self.init_vizrank()

    @OWAnchorProjectionWidget.Inputs.data
    def set_data(self, data):
        super().set_data(data)
        self.init_vizrank()
        self.init_projection()

    def init_vizrank(self):
        msgerr = ""
        if self.data is None:
            msgerr = "No data"
        elif len(self.primitive_variables) <= 3:
            msgerr = "Not enough variables"
        elif self.attr_color is None:
            msgerr = "Color is not set."
        elif np.isnan(self.data.get_column(self.attr_color)).all():
            msgerr = "No rows with defined color variable"
        elif np.sum(np.all(np.isfinite(self.data.X), axis=1)) <= 1:
            msgerr = "Not enough rows without missing data"
        elif not np.all(np.nan_to_num(np.nanstd(self.data.X, 0)) != 0):
            msgerr = "Constant data"

        if not msgerr:
            super().init_vizrank(
                self.data, self.primitive_variables, self.attr_color,
                self.n_attrs_vizrank)
        else:
            self.disable_vizrank(msgerr)

    def check_data(self):
        super().check_data()
        if self.data is not None:
            domain = self.data.domain
            vars_ = chain(domain.variables, domain.metas)
            n_vars = sum(v.is_primitive() for v in vars_)
            if len(self.primitive_variables) < n_vars:
                self.Warning.removed_vars()

    def init_attr_values(self):
        super().init_attr_values()
        self.selected_vars[:] = self.primitive_variables[:5]
        self.model_selected[:] = self.primitive_variables

    def _manual_move(self, anchor_idx, x, y):
        angle = np.arctan2(y, x)
        super()._manual_move(anchor_idx, np.cos(angle), np.sin(angle))

    def _send_components_x(self):
        components_ = super()._send_components_x()
        angle = np.arctan2(*components_[::-1])
        return np.row_stack((components_, angle))

    def _send_components_metas(self):
        return np.vstack((super()._send_components_metas(), ["angle"]))

    def clear(self):
        super().clear()
        self.projector = RadViz()

    @classmethod
    def migrate_context(cls, context, version):
        values = context.values
        if version < 2:
            values["attr_color"] = values["graph"]["attr_color"]
            values["attr_size"] = values["graph"]["attr_size"]
            values["attr_shape"] = values["graph"]["attr_shape"]
            values["attr_label"] = values["graph"]["attr_label"]
        if version < 3 and "selected_vars" in values:
            values["selected_vars"] = (values["selected_vars"], -3)


class MoveIndicator(pg.GraphicsObject):
    def __init__(self, angle, col, dangle=5, parent=None):
        super().__init__(parent)
        color = QColor(0, 0, 0) if col else QColor(128, 128, 128)
        angle_d = np.rad2deg(angle)
        angle_2 = 90 - angle_d - dangle
        angle_1 = 270 - angle_d + dangle
        dangle = np.deg2rad(dangle)
        arrow1 = pg.ArrowItem(
            parent=self, angle=angle_1, brush=color, pen=pg.mkPen(color)
        )
        arrow1.setPos(np.cos(angle - dangle), np.sin(angle - dangle))
        arrow2 = pg.ArrowItem(
            parent=self, angle=angle_2, brush=color, pen=pg.mkPen(color)
        )
        arrow2.setPos(np.cos(angle + dangle), np.sin(angle + dangle))
        arc_x = np.fromfunction(
            lambda i: np.cos((angle - dangle) + (2 * dangle) * i / 120.),
            (121,), dtype=int
        )
        arc_y = np.fromfunction(
            lambda i: np.sin((angle - dangle) + (2 * dangle) * i / 120.),
            (121,), dtype=int
        )
        pg.PlotCurveItem(
            parent=self, x=arc_x, y=arc_y, pen=pg.mkPen(color), antialias=False
        )

    def paint(self, painter, option, widget):
        pass

    def boundingRect(self):
        return QRectF()


if __name__ == "__main__":  # pragma: no cover
    brown = Table("brown-selected")
    WidgetPreview(OWRadviz).run(set_data=brown, set_subset_data=brown[::10])


import unittest
from unittest.mock import patch, Mock

from AnyQt.QtCore import Qt
from AnyQt.QtGui import QStandardItem
from AnyQt.QtWidgets import QDialog
from AnyQt.QtTest import QSignalSpy

from orangewidget.tests.base import GuiTest
from Orange.widgets.widget import OWWidget
from Orange.data import Table, Domain, ContinuousVariable
from Orange.widgets.visualize.utils.vizrank import (
    RunState, VizRankMixin,
    VizRankDialog, VizRankDialogAttrs, VizRankDialogAttrPair,
    VizRankDialogNAttrs
)


class MockDialog(VizRankDialog):
    class Task:
        interrupt = False
        set_partial_result = Mock()

        def is_interruption_requested(self):
            return self.interrupt

    task = Task()

    def __init__(self, parent=None):
        super().__init__(parent)

    # False positive?! "Number of parameters was 4 in 'ConcurrentMixin.start'
    # and is now 4 in overriding 'MockDialog.start' method".
    # pylint: disable=arguments-differ
    def start(self, func, *args, **kwargs):
        func(*args, self.task)

    def state_generator(self):
        return range(10)

    def state_count(self):
        return 10

    def compute_score(self, state):
        return 10 * (state % 2) + state // 2 if state != 3 else None

    row_for_state = Mock()


class TestVizRankDialog(GuiTest):

    @patch.object(VizRankDialog, "start")
    @patch.object(VizRankDialog, "prepare_run",
                  new=lambda self: setattr(self.run_state,
                                           "state", RunState.Ready))
    def test_init_and_button(self, run_vizrank):
        dialog = VizRankDialog(None)
        self.assertEqual(dialog.run_state.state, RunState.Initialized)
        self.assertEqual(dialog.button.text(), dialog.button_labels[RunState.Initialized])
        dialog.button.click()
        run_vizrank.assert_called_once()
        self.assertEqual(dialog.run_state.state, RunState.Running)
        self.assertEqual(dialog.button.text(), dialog.button_labels[RunState.Running])

        dialog.button.click()
        run_vizrank.assert_called_once()
        self.assertEqual(dialog.run_state.state, RunState.Paused)
        self.assertEqual(dialog.button.text(), dialog.button_labels[RunState.Paused])

        dialog.button.click()
        self.assertEqual(run_vizrank.call_count, 2)
        self.assertEqual(dialog.run_state.state, RunState.Running)
        self.assertEqual(dialog.button.text(), dialog.button_labels[RunState.Running])

    def test_running(self):
        dialog = MockDialog()
        dialog.task.interrupt = False
        dialog.start_computation()
        result = dialog.task.set_partial_result.call_args[0][0]
        self.assertEqual(result.scores, [0, 1, 2, 3, 4, 10, 12, 13, 14])
        self.assertEqual(result.completed_states, 10)

        dialog.rank_model.appendRow(QStandardItem("foo"))
        dialog.rank_model.appendRow(QStandardItem("bar"))
        dialog.on_done(result)
        self.assertEqual(dialog.run_state.state, RunState.Done)
        self.assertEqual(dialog.button.text(), dialog.button_labels[RunState.Done])
        self.assertFalse(dialog.button.isEnabled())
        selection = dialog.rank_table.selectedIndexes()
        self.assertEqual(len(selection), 1)
        self.assertEqual(selection[0].row(), 0)

    def test_interruption(self):
        dialog = MockDialog()
        dialog.task.interrupt = True
        dialog.start_computation()
        result = dialog.task.set_partial_result.call_args[0][0]
        self.assertEqual(result.scores, [0])
        self.assertEqual(result.completed_states, 1)


class TestVizRankMixin(GuiTest):
    def setUp(self):
        self.mock_dialog = Mock()
        self.mock_dialog.__name__ = "foo"

        class Widget(OWWidget, VizRankMixin(self.mock_dialog)):
            pass

        self.widget = Widget()

    def test_button(self):
        widget, dialog = self.widget, self.mock_dialog

        # False positives, pylint: disable=attribute-defined-outside-init)
        widget.start_vizrank = Mock()
        widget.raise_vizrank = Mock()

        button = widget.vizrank_button("Let's Vizrank")
        self.assertEqual(button.text(), "Let's Vizrank")
        self.assertFalse(button.isEnabled())

        widget.disable_vizrank("Too lazy to rank to-day.")
        self.assertEqual(button.text(), "Let's Vizrank")
        self.assertEqual(button.toolTip(), "Too lazy to rank to-day.")
        self.assertFalse(button.isEnabled())
        dialog.assert_not_called()

        widget.init_vizrank()
        dialog.assert_called_once()
        dialog.reset_mock()
        self.assertEqual(button.text(), "Let's Vizrank")
        self.assertEqual(button.toolTip(), "")
        self.assertTrue(button.isEnabled())

        widget.disable_vizrank("Too lazy to rank to-day.")
        self.assertEqual(button.text(), "Let's Vizrank")
        self.assertEqual(button.toolTip(), "Too lazy to rank to-day.")
        self.assertFalse(button.isEnabled())
        dialog.assert_not_called()

        widget.init_vizrank()
        dialog.assert_called_once()  # new data, new dialog!
        dialog.reset_mock()

        button.click()
        widget.start_vizrank.assert_called_once()
        widget.raise_vizrank.assert_called_once()

    def test_no_button(self):
        widget, dialog = self.widget, self.mock_dialog

        widget.disable_vizrank("Too lazy to rank to-day.")
        dialog.assert_not_called()

        widget.init_vizrank()
        dialog.assert_called_once()
        dialog.reset_mock()
        widget.disable_vizrank("Too lazy to rank to-day.")

        widget.disable_vizrank("Too lazy to rank to-day.")
        dialog.assert_not_called()

        widget.init_vizrank()
        dialog.assert_called_once()
        dialog.reset_mock()
        widget.disable_vizrank("Too lazy to rank to-day.")

    def test_init_vizrank(self):
        widget, dialog = self.widget, self.mock_dialog
        a, b = Mock(), Mock()
        widget.init_vizrank(a, b)
        dialog.assert_called_with(widget, a, b)


class TestVizRankDialogWithData(GuiTest):
    def setUp(self):
        self.attrs = tuple(ContinuousVariable(n) for n in "abcdef")
        self.class_var = ContinuousVariable("y")
        self.variables = (*self.attrs, self.class_var)
        self.metas = tuple(ContinuousVariable(n) for n in "mn")
        self.data = Table.from_list(
            Domain(self.attrs, self.class_var, self.metas),
            [[0] * 9])


class TestVizRankDialogAttrs(TestVizRankDialogWithData):
    def test_init(self):
        dialog = VizRankDialogAttrs(None, self.data)
        self.assertIs(dialog.data, self.data)
        self.assertEqual(dialog.attrs, self.variables)
        self.assertIsNone(dialog.attr_color)

        dialog = VizRankDialogAttrs(None, self.data, self.attrs[1:])
        self.assertIs(dialog.data, self.data)
        self.assertEqual(dialog.attrs, self.attrs[1:])
        self.assertIsNone(dialog.attr_color)

        dialog = VizRankDialogAttrs(None, self.data, self.attrs[1:], self.attrs[3])
        self.assertIs(dialog.data, self.data)
        self.assertEqual(dialog.attrs, self.attrs[1:])
        self.assertIs(dialog.attr_color, self.attrs[3])

    def test_attr_order(self):
        # pylint: disable=abstract-method
        dialog = VizRankDialogAttrs(None, self.data)
        self.assertEqual(dialog.attr_order, self.variables)

        class OrderedAttr(VizRankDialogAttrs):
            call_count = 0

            def score_attributes(self):
                self.call_count += 1
                return self.attrs[::-1]

        dialog = OrderedAttr(None, self.data)
        self.assertEqual(dialog.attr_order, self.variables[::-1])
        self.assertEqual(dialog.attr_order, self.variables[::-1])
        self.assertEqual(dialog.call_count, 1)

    def test_row_for_state(self):
        # pylint: disable=protected-access
        dialog = VizRankDialogAttrs(None, self.data)
        item = dialog.row_for_state(0, [3, 1])[0]
        self.assertEqual(item.data(Qt.DisplayRole), "d, b")
        self.assertEqual(item.data(dialog._AttrRole), [self.attrs[3], self.attrs[1]])

        dialog.sort_names_in_row = True
        item = dialog.row_for_state(0, [3, 1])[0]
        self.assertEqual(item.data(Qt.DisplayRole), "b, d")
        self.assertEqual(item.data(dialog._AttrRole), [self.attrs[1], self.attrs[3]])

    def test_autoselect(self):
        # False positive, pylint: disable=no-value-for-parameter
        class Widget(OWWidget, VizRankMixin(VizRankDialogAttrs)):
            pass
        widget = Widget()
        try:
            widget.init_vizrank(self.data)
            dialog = widget.vizrank_dialog
            for state in ([0, 1], [0, 3], [1, 3], [3, 1]):
                dialog.rank_model.appendRow(dialog.row_for_state(0, state))

            widget.vizrankAutoSelect.emit([self.attrs[1], self.attrs[3]])
            selection = dialog.rank_table.selectedIndexes()
            self.assertEqual(len(selection), 1)
            self.assertEqual(selection[0].row(), 2)

            widget.vizrankAutoSelect.emit([self.attrs[3], self.attrs[1]])
            selection = dialog.rank_table.selectedIndexes()
            self.assertEqual(len(selection), 1)
            self.assertEqual(selection[0].row(), 3)

            widget.vizrankAutoSelect.emit([self.attrs[3], self.attrs[0]])
            selection = dialog.rank_table.selectedIndexes()
            self.assertEqual(len(selection), 0)
        finally:
            widget.onDeleteWidget()


class TestVizRankDialogAttrPair(TestVizRankDialogWithData):
    def test_count_and_generator(self):
        dialog = VizRankDialogAttrPair(None, self.data, self.attrs[:5])
        self.assertEqual(dialog.state_count(), 5 * 4 / 2)
        self.assertEqual(
            list(dialog.state_generator()),
            [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3),
             (0, 4), (1, 4), (2, 4), (3, 4)])


class TestVizRankDialogNAttrs(TestVizRankDialogWithData):
    def test_spin_interaction(self):
        dialog = VizRankDialogNAttrs(None,
                                     self.data, self.attrs[:5], None, 4)
        spy = QSignalSpy(dialog.runStateChanged)

        with patch.object(
                VizRankDialog, "pause_computation",
                side_effect=lambda: dialog.set_run_state(RunState.Paused)):
            with patch.object(
                    VizRankDialog, "start_computation",
                    side_effect=lambda: dialog.set_run_state(RunState.Running)):
                spin = dialog.n_attrs_spin
                self.assertEqual(spin.value(), 4)
                self.assertEqual(spin.maximum(), 5)

                dialog.start_computation()
                self.assertEqual(dialog.run_state.state, RunState.Running)
                self.assertEqual(spy[-1][1]["n_attrs"], 4)

                spin.setValue(3)
                # Ranking must be paused
                self.assertEqual(dialog.run_state.state, RunState.Paused)
                # Label should be changed to "restart with ..."
                self.assertNotEqual(dialog.button.text(),
                                    dialog.button_labels[RunState.Paused])

                spin.setValue(4)
                self.assertEqual(dialog.run_state.state, RunState.Paused)
                # Label should be reset to "Continue"
                self.assertEqual(dialog.button.text(),
                                 dialog.button_labels[RunState.Paused])

        # Remove the side-effect so that we see that start_computation
        # resets the state to Initialized before calling super
        with patch.object(VizRankDialog, "start_computation"):
            dialog.start_computation()
            self.assertEqual(spy[-1][1]["n_attrs"], 4)
            # Here, the state must not be reset to Initialized
            self.assertEqual(dialog.run_state.state, RunState.Paused)
            # But now manually set it to appropriate state
            dialog.set_run_state(RunState.Running)

            spin.setValue(3)
            self.assertEqual(dialog.run_state.state, RunState.Paused)
            self.assertNotEqual(dialog.button.text(), dialog.button_labels[RunState.Paused])

            dialog.start_computation()
            self.assertEqual(dialog.run_state.state, RunState.Initialized)
            self.assertEqual(spy[-1][1]["n_attrs"], 3)


if __name__ == "__main__":
    unittest.main()




# THIS FILE IS GENERATED FROM ORANGE SETUP.PY
short_version = '3.38.5'
version = '3.38.5'
full_version = '3.38.5.dev0+2fbbf2c'
git_revision = '2fbbf2cc8436e6799de543bd2758774776acb4e5'
release = False

if not release:
    version = full_version
